import unittest
from collections import deque
from ..event import Event, EventQueue
from ..monad import Cont

__all__ = ('EvetTest', 'EventQueueTest',)


class EventTest(unittest.TestCase):
    def test(self):
        def handler(tag, ret):
            def handler(val):
                queue.append((tag, val))
                return ret
            return handler
        queue = deque()

        event = Event()
        event(0)

        h1 = event.on(handler('1', False))
        h2 = event.on(handler('2', True))

        event(1)
        self.assertEqual(queue.popleft(), ('1', 1))
        self.assertEqual(queue.popleft(), ('2', 1))
        self.assertFalse(queue)

        event(2)
        self.assertEqual(queue.popleft(), ('2', 2))
        self.assertFalse(queue)
        self.assertEqual(event.off(h1), False)
        self.assertEqual(event.off(h2), True)

    def test_monad(self):
        ret = lambda tag: lambda val: queue.append((tag, val))
        queue = deque()

        event = Event()
        event(0)

        monad = event.__monad__()
        self.assertTrue(isinstance(monad, Cont))
        monad(ret('1'))
        monad(ret('2'))

        event(1)
        self.assertEqual(queue.popleft(), ('1', 1))
        self.assertEqual(queue.popleft(), ('2', 1))
        self.assertFalse(queue)

        event(2)
        self.assertFalse(queue)
        self.assertFalse(event.handlers)

    def test_handler(self):
        rets = []
        event = Event()

        def h0(val):
            event('done')
            event.on(h1)

        def h1(val):
            rets.append(val)

        event.on(h0)
        event(0)
        self.assertEqual(rets, [])

        event(1)
        self.assertEqual(rets, [1])

    def test_reduce(self):
        items = []
        get = lambda: event.__monad__()(lambda val: items.append(val))
        event = Event()

        with event.reduce() as revent:
            get()
            self.assertEqual(items, [])

            revent(0)
            self.assertEqual(items, [0])

            revent(1)
            self.assertEqual(items, [0])

        with self.assertRaises(ValueError):
            revent(2)


class EventQueueTest(unittest.TestCase):
    def test(self):
        items = []
        get = lambda: queue.__monad__()(lambda val: items.append(val))
        queue = EventQueue()

        get()
        self.assertEqual(items, [])
        self.assertEqual(len(queue), 0)

        queue(0)
        self.assertEqual(items, [0])
        self.assertEqual(len(queue), 0)

        queue(1)
        queue(2)
        self.assertEqual(items, [0])
        self.assertEqual(len(queue), 2)

        get()
        self.assertEqual(items, [0, 1])
        self.assertEqual(len(queue), 1)

        get()
        self.assertEqual(items, [0, 1, 2])
        self.assertEqual(len(queue), 0)



"""Base monad type with implementation of common monadic functions.
"""
from .proxy import Proxy
from functools import wraps, reduce

__all__ = ('Monad',)


class Monad(object):
    """Base monad type
    """
    __slots__ = tuple()

    def __init__(self):
        raise NotImplementedError()

    def __monad__(self):
        return self

    @classmethod
    def unit(cls, val):
        """Create primitive monad from value

        unit :: (Monad m) => a -> m a
        """
        raise NotImplementedError()

    def bind(self, func):
        """Bind function

        bind :: (Monad m) => m a -> (a -> m b) -> m b
        """
        raise NotImplementedError()

    @property
    def value(self):
        """Get embedded monad value

        This monad independent version works only inside do_green block.
        """
        from .do_green import bind_green
        return bind_green(self)

    @property
    def proxy(self):
        """Put this monad inside proxy monad
        """
        return Proxy(self)

    def __irshift__(self, func):
        return self.bind(func)

    @classmethod
    def zero(cls):
        """Zero monad

        zero :: (Monad m) => () -> m a
        """
        return NotImplementedError()

    def plus(self, monad):
        """Monad plus operation

        plus :: (Monad m) => m a -> m b -> m c
        Plus laws:
            (m0 + m1) + m2 == m0 + (m1 + m2)
            m + zero == m
        """
        return NotImplementedError()

    def __add__(self, monad):
        return self.plus(monad)

    def then(self, monad):
        return self.bind(lambda _: monad)

    def __rshift__(self, monad):
        return self.then(monad)

    def then_val(self, val):
        return self.then(self.unit(val))

    def map_val(self, func):
        """Functors map operation

        map :: (Monad m) => m a -> (a -> b) -> m b
        """
        return self.bind(lambda val: self.unit(func(val)))

    def join(self):
        """Join nested monad

        The 'join' function is the conventional monad join operator. It is used
        to remove one level of monadic structure, projecting its bound argument
        into the outer level.

        join :: (Monad m) => m (m a) -> m a
        """
        return self.bind(lambda m: m.__monad__())

    @classmethod
    def map(Monad, mapper, vals):
        """Map with monadic function (mapM)

        map :: (Monad m) => (a -> m b) -> [a] -> m [b]
        """
        return Monad.sequence(mapper(val) for val in vals)

    @classmethod
    def filter(Monad, pred, vals):
        """Filter with monadic predicate (filterM)

        filter :: (Monad m) => (a -> m Bool) -> [a] -> m [a]
        """
        def chain(mvals, val):
            return mvals.bind(lambda vals: pred(val).__monad__().bind(
                              lambda add: Monad.unit(vals + (val,) if add else vals)))
        return reduce(chain, vals, Monad.unit(tuple()))

    @classmethod
    def fold(Monad, folder, acc, vals):
        """Fold with monadic function (foldM)

        fold :: (Monad m) => (a -> b -> m a) -> a -> [b] -> m a
        """
        def fold(acc, vals):
            if not vals:
                return Monad.unit(acc)
            val, vals = vals[0], vals[1:]
            return folder(acc, val).__monad__().bind(lambda acc: fold(acc, vals))
        vals = tuple(vals)
        return fold(acc, vals)

    @classmethod
    def sequence(Monad, monads):
        """Sequence monads

        Evaluate each action in the sequence from left to right, and collect the results.

        sequence :: Monad m => [m a] -> m [a]
        """
        def chain(mvals, mval):
            return mvals.bind(lambda vals: mval.bind(lambda val: Monad.unit(vals + (val,))))
        return reduce(chain, (monad.__monad__() for monad in monads), Monad.unit(tuple()))

    def __and__(self, monad):
        """Compose two monads into one

        Equivalent to sequence((self, monad)).
        (&) :: m a -> m b -> m (a, b)
        """
        return self.bind(lambda val: monad.__monad__().bind(
                         lambda mval: self.unit((val, mval))))

    @classmethod
    def lift(Monad, monad):
        """Wrap monad's value inside primitive monad of this type
        """
        monad = monad.__monad__()
        return monad.bind(lambda val: monad.unit(Monad.unit(val)))

    @classmethod
    def lift_func(Monad, func):
        """Promote function to fully monadic one (arguments and result)

        lift_func :: Monad m => (a0 -> ... -> aN -> r) -> (m a0 -> ... -> m aN -> m r)
        """
        return wraps(func)(lambda *margs: Monad.sequence(margs).bind(lambda args: Monad.unit(func(*args))))

    @classmethod
    def lift_func_result(Monad, func):
        """Promote function to monadic one (result only)

        lift_func_result :: Monad
        """
        return wraps(func)(lambda *args, **kw: Monad.unit(func(*args, **kw)))

    @classmethod
    def ap(Monad, mfunc):
        """Applicative <*> operator

        ap :: m (a -> b) -> m a -> m b
        """
        return lambda ma: mfunc.bind(lambda func: ma.bind(lambda a: Monad.unit(func(a))))



"""SSH Connection

Connection with remote machine over ssh protocol.
"""
from .shell import ShellConnection

__all__ = ('SSHConnection',)


class SSHConnection(ShellConnection):
    """SSH Connection
    """
    def __init__(self, host, port=None, identity_file=None, ssh_exec=None,
                 py_exec=None, environ=None, bufsize=None, hub=None, core=None):
        self.host = host
        self.port = port
        self.identity_file = identity_file
        self.ssh_exec = ssh_exec or 'ssh'

        # ssh command
        command = [
            self.ssh_exec,          # command
            '-T',                   # disable pseudo-tty allocation
            '-o', 'BatchMode=yes',  # never ask password
            self.host,              # host
        ]
        command.extend(('-i', self.identity_file) if self.identity_file else [])
        command.extend(('-p', self.port) if self.port else [])

        ShellConnection.__init__(self, command=command, escape=True,
                                 py_exec=py_exec, environ=environ,
                                 bufsize=bufsize, hub=hub, core=core)



import unittest
from ..utils import lazy

__all__ = ('LazyTest',)


class LazyTest(unittest.TestCase):
    def test(self):
        @lazy
        def lazy_val():
            calls[0] += 1
            return 'value'
        calls = [0]
        self.assertEqual(calls, [0])

        self.assertEqual(lazy_val(), 'value') 
        self.assertEqual(calls, [1])

        self.assertEqual(lazy_val(), 'value')
        self.assertEqual(calls, [1])




"""File stream

File stream is stream created from file descriptor.
"""
import os
import errno
import fcntl
from .stream import Stream
from .buffered import BufferedStream
from ..core import Core, POLL_READ, POLL_WRITE
from ..monad import async, do_return
from ..uniform import BrokenPipeError, BlockingErrorSet, PipeErrorSet

__all__ = ('File', 'BufferedFile', 'fd_close_on_exec', 'fd_blocking',)


class File(Stream):
    """File stream

    File stream is stream created from file descriptor.
    """
    def __init__(self, fd, closefd=None, init=None, core=None):
        Stream.__init__(self)

        self.fd = fd if isinstance(fd, int) else fd.fileno()
        self.closefd = closefd is None or closefd
        self.core = core or Core.local()
        self.blocking(False)
        if init is None or init:
            self.initing()

    def fileno(self):
        return self.fd

    @async
    def read(self, size):
        with self.reading:
            while True:
                try:
                    data = os.read(self.fd, size)
                    if size and not data:
                        raise BrokenPipeError(errno.EPIPE, 'broken pipe')
                    do_return(data)
                except OSError as error:
                    if error.errno not in BlockingErrorSet:
                        if error.errno in PipeErrorSet:
                            raise BrokenPipeError(error.errno, error.strerror)
                        raise
                yield self.core.poll(self.fd, POLL_READ)

    @async
    def write(self, data):
        with self.writing:
            while True:
                try:
                    do_return(os.write(self.fd, data))
                except OSError as error:
                    if error.errno not in BlockingErrorSet:
                        if error.errno in PipeErrorSet:
                            raise BrokenPipeError(error.errno, error.strerror)
                        raise
                yield self.core.poll(self.fd, POLL_WRITE)

    def dispose(self):
        if Stream.dispose(self):
            fd, self.fd = self.fd, -1
            self.core.poll(fd, None)()  # resolve with BrokenPipeError
            if self.closefd:
                os.close(fd)
            return True
        return False

    def detach(self):
        if self.disposed:
            raise ValueError('file is disposed')
        try:
            self.closefd = False
            self.blocking(True)
            return self.fd
        finally:
            self.dispose()

    def blocking(self, enable=None):
        return fd_blocking(self.fd, enable)

    def close_on_exec(self, enable=None):
        return fd_close_on_exec(self.fd, enable)

    def __str__(self):
        flags = []
        if self.fd > 0:
            self.blocking() and flags.append('blocking')
            self.close_on_exec() and flags.append('close_on_exec')
        return ('{}(fd:{}, flags:{}, state:{})'.format(type(self).__name__,
                self.fd, ','.join(flags), self.state.state_name()))

    def __repr__(self):
        return str(self)


class BufferedFile(BufferedStream):
    def __init__(self, fd, bufsize=None, closefd=None, core=None):
        BufferedStream.__init__(self, File(fd, closefd, True, core), bufsize)

    def detach(self):
        return BufferedStream.detach(self).detach()


def fd_close_on_exec(fd, enable=None):
    """Set or get file descriptors close_on_exec flag
    """
    return fd_option(fd, fcntl.F_GETFD, fcntl.F_SETFD, fcntl.FD_CLOEXEC, enable)


def fd_blocking(fd, enable=None):
    """Set or get file descriptors blocking flag
    """
    return not fd_option(fd, fcntl.F_GETFL, fcntl.F_SETFL, os.O_NONBLOCK,
                         None if enable is None else not enable)


def fd_option(fd, get_flag, set_flag, option_flag, enable=None):
    """Set or get file descriptor option
    """
    options = fcntl.fcntl(fd, get_flag)
    if enable is None:
        return bool(options & option_flag)
    elif enable:
        options |= option_flag
    else:
        options &= ~option_flag
    fcntl.fcntl(fd, set_flag, options)
    return enable



"""Remote importer
"""
import os
import sys
import socket
import inspect
import pkgutil
from .hub import pair
from .expr import SetItemExpr, GetAttrExpr, LoadArgExpr
from ..core import Core
from ..monad import Result, async, do_return
from ..boot import BootLoader

__all__ = ('Importer',)


class Importer(object):
    def __init__(self, sender, location=None):
        self.sender = sender
        self.location = location or (socket.gethostname(), os.getpid())
        self.loaders = {}

    @classmethod
    def create(cls, hub=None):
        """Create importer proxy object
        """
        def importer_handler(name, dst, src):
            try:
                if name is None:
                    return False  # dispose importer
                module = sys.modules.get(name, False)
                if module is None:
                    src.send(None)  # Module is cached as not found (python 2)
                    return True
                loader = pkgutil.get_loader(name)
                if loader is None or not hasattr(loader, 'get_source'):
                    src.send(None)
                    return True
                source = loader.get_source(name)
                if source is None:
                    src.send(None)
                    return True
                ispkg = loader.is_package(name)
                if module and hasattr(module, '__package__'):
                    pkg = module.__package__
                else:
                    pkg = name if ispkg else name.rpartition('.')[0]
                try:
                    filename = (inspect.getfile(loader.get_code(name)) if not module else
                                inspect.getfile(module))
                except TypeError:
                    filename = '<unknown>'
                src.send(BootLoader(name, source, filename, ispkg, pkg))
            except Exception:
                src.send(Result.from_current_error())
            return True

        recv, send = pair(hub=hub)
        recv(importer_handler)
        return Importer(send)

    @classmethod
    @async
    def create_remote(cls, conn, index=None):
        """Create and install importer on specified connection
        """
        importer = Importer.create(conn.hub)
        try:
            yield conn(importer)(index)

            # determine full name of __main__ module
            module = sys.modules.get('__main__', None)
            while module is not None:
                try:
                    file = inspect.getsourcefile(module)
                    if not file:
                        break
                except TypeError:
                    # __main__ is a built-in module. Do not need to map anything
                    break
                name = os.path.basename(file).partition('.')[0]
                package = getattr(module, '__package__', None)
                if package:
                    name = '{}.{}'.format(package, name)
                else:
                    try:
                        source = inspect.getsource(module)
                    except Exception:
                        # __main__ source file is <stdin> or something like this
                        break
                    loader = BootLoader(name, source, file, False, None)
                    yield conn(loader)().__package__

                # remote_conn.module_map['__main__'] = main
                yield conn.sender(SetItemExpr(GetAttrExpr(LoadArgExpr(0),
                                  'module_map'), '__main__', name).code())
                conn.module_map[name] = '__main__'
                break
            do_return(importer)

        except Exception:
            importer.dispose()
            raise

    def __call__(self, index=None):
        """Install importer
        """
        if self in sys.meta_path:
            return
        if index is None:
            sys.meta_path.append(self)
        else:
            sys.meta_path.insert(index, self)

    def find_module(self, name, path=None):
        if self.sender is None:
            return None
        loader = self.loaders.get(name, False)
        if loader is False:
            # Function find_module must be synchronous, so we must execute core
            # until request is fulfilled.
            loader = self.sender(name).future()
            for _ in Core.local():
                if loader.completed:
                    loader = loader.value
                    break
            self.loaders[name] = loader
        if loader is None:
            return None
        return loader

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.sender == other.sender

    def __hash__(self):
        return hash(self.sender)

    def __reduce__(self):
        return Importer, (self.sender, self.location)

    def dispose(self):
        if self in sys.meta_path:
            sys.meta_path.remove(self)
        sender, self.sender = self.sender, None
        if sender is not None:
            try:
                sender.send(None)
            except ValueError:
                pass

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return ('{}(addr:{}, host:{}, pid:{})'.format(type(self).__name__,
                self.sender.addr if self.sender else None,
                *self.location))

    def __repr__(self):
        return str(self)



"""Thread pool
"""
import os
import threading
import multiprocessing
from collections import deque
from .core import Core
from .monad import Result, async_block

__all__ = ('ThreadPool', 'task',)


class ThreadPool(object):
    """Thread pool
    """
    inst_lock = threading.RLock()
    inst_main = None

    def __init__(self, size=None):
        self.disposed = False

        self.threads = set()
        self.thread_count = size or _cpu_count()
        self.thread_idle = 0
        self.thread_queue = deque()
        self.thread_lock = threading.RLock()
        self.thread_cond = threading.Condition(self.thread_lock)

    @classmethod
    def main(cls, inst=None):
        """Access or create main thread pool object
        """
        with cls.inst_lock:
            if inst is None:
                if cls.inst_main is None:
                    cls.inst_main = cls()
                inst = cls.inst_main
            else:
                cls.inst_main = inst
        return inst

    @property
    def size(self):
        """Size of thread pool (number of threads)
        """
        return self.thread_count

    @size.setter
    def size(self, count):
        """Change size of thread pool (number of threads)
        """
        if count < 0:
            raise ValueError('thread pool size must be positive')
        with self.thread_lock:
            self.thread_count = count
            exit_count = len(self.threads) - count
            if exit_count > 0:
                for _ in range(exit_count):
                    self.thread_queue.appendleft(_action_exit)
                self.thread_cond.notify(exit_count)

    def __len__(self):
        return self.size

    def __call__(self, act, *act_a, **act_kw):
        """Schedule action to be executed on main thread pool

        Returns continuation which will be resolved with action result on
        local core object.
        """
        if self.disposed:
            raise ValueError('thread pool is disposed')

        @async_block
        def thread_cont(ret):
            def action(error=None):
                if error is not None:
                    res = Result.from_exception(error)
                else:
                    try:
                        res = act(*act_a, **act_kw)
                    except Exception:
                        res = Result.from_current_error()
                core.schedule()(lambda _: ret(res))
            core = Core.local()  # capture caller's core object

            with self.thread_lock:
                self.thread_queue.append(action)
                if not self.thread_idle and len(self.threads) < self.thread_count:
                    thread = threading.Thread(target=self.thread_main)
                    thread.daemon = True
                    self.threads.add(thread)
                    thread.start()
                else:
                    self.thread_cond.notify()

        return thread_cont

    def thread_main(self):
        """Worker main function
        """
        with self.thread_lock:
            if threading.current_thread() not in self.threads:
                raise ValueError('this function must be run in thread pool')
        Core.local(_bad_object(RuntimeError('can not access core inside '
                                            'thread pool\'s worker thread')))
        try:
            while True:
                with self.thread_lock:
                    while not self.thread_queue:
                        self.thread_idle += 1
                        self.thread_cond.wait()
                        self.thread_idle -= 1
                    action = self.thread_queue.popleft()
                action()
        except _thread_exit:
            pass
        finally:
            with self.thread_lock:
                self.threads.discard(threading.current_thread())

    def dispose(self):
        with self.thread_lock:
            if self.disposed:
                return
            elif threading.current_thread() in self.threads:
                raise ValueError('thread pool cannot be disposed from its own thread')
            self.disposed = True
            actions, self.thread_queue = self.thread_queue, deque((_action_exit,) * self.thread_count)
            self.thread_cond.notify(self.thread_count)
        for action in actions:
            action(ValueError('thread pool has been disposed'))

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return ('ThreadPool(size:{}, threads:{}, idle:{}, tasks:{})'.format(
                self.size, len(self.threads), self.thread_idle,
                len(self.thread_queue) + len(self.threads) - self.thread_idle))

    def __repr__(self):
        return str(self)


def task(action, *action_a, **action_kw):
    """Schedule action to be executed on main thread pool

    Returns continuation which will be resolved with action result on
    local core object.
    """
    return ThreadPool.main()(action, *action_a, **action_kw)


class _thread_exit(Exception):
    """Helper thread exit exception
    """


def _action_exit(error=None):
    if error is None:
        raise _thread_exit()


class _bad_object(object):
    """Bad object

    Raises and error on attempt of attribute access or call.
    """
    def __init__(self, error):
        object.__setattr__(self, '_error', error)

    def __getattr__(self, name):
        raise self._error

    def __setattr__(self, name, value):
        raise self._error

    def __call__(self, *a, **kw):
        raise self._error


def _cpu_count():
    """Returns the number of processors on this machine."""
    try:
        return multiprocessing.cpu_count()
    except NotImplementedError:
        pass
    try:
        return os.sysconf("SC_NPROCESSORS_CONF")
    except ValueError:
        pass
    return 1




__all__ = []


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import hub, expr, proxy, conn, closure, composite

    suite = TestSuite()
    for test in (hub, expr, proxy, conn, closure, composite):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite



"""Buffered stream

Buffered stream is wrapped stream. Implements additional methods which leverage
buffered nature of the stream.
"""
import struct
from collections import deque
from .wrapped import WrappedStream
from .. import PRETZEL_BUFSIZE
from ..uniform import BrokenPipeError
from ..monad import async, async_single, do_return

__all__ = ('BufferedStream',)


class BufferedStream(WrappedStream):
    """Buffered stream
    """
    size_struct = struct.Struct('>I')

    def __init__(self, base, bufsize=None):
        WrappedStream.__init__(self, base)

        self.bufsize = bufsize or PRETZEL_BUFSIZE
        self.read_buffer = Buffer()
        self.write_buffer = Buffer()

        @async_single
        @async
        def flush():
            """Flush write buffers
            """
            with self.writing:
                while self.write_buffer:
                    block = self.write_buffer.slice(self.bufsize)
                    self.write_buffer.dequeue((yield self.base.write(block)), False)
                yield self.base.flush()
        self.flush = flush

    @async
    def read(self, size):
        if not size:
            do_return(b'')
        with self.reading:
            if not self.read_buffer:
                self.read_buffer.enqueue((yield self.base.read(self.bufsize)))
            do_return(self.read_buffer.dequeue(size))

    @async
    def read_until_size(self, size):
        """Read exactly size bytes
        """
        if not size:
            do_return(b'')
        with self.reading:
            while len(self.read_buffer) < size:
                self.read_buffer.enqueue((yield self.base.read(self.bufsize)))
            do_return(self.read_buffer.dequeue(size))

    @async
    def read_until_eof(self):
        """Read until stream is closed
        """
        with self.reading:
            try:
                while True:
                    self.read_buffer.enqueue((yield self.base.read(self.bufsize)))
            except BrokenPipeError:
                pass
            do_return(self.read_buffer.dequeue())

    @async
    def read_until_sub(self, sub=None):
        """Read until substring is found

        Returns data including substring. Default substring is "\\n".
        """
        sub = sub or b'\n'
        with self.reading:
            offset = 0
            while True:
                data = self.read_buffer.slice()
                find_offset = data[offset:].find(sub)
                if find_offset >= 0:
                    break
                offset = max(0, len(data) - len(sub))
                self.read_buffer.enqueue((yield self.base.read(self.bufsize)))
            do_return(self.read_buffer.dequeue(offset + find_offset + len(sub)))

    @async
    def read_until_regex(self, regex):
        """Read until regular expression is matched

        Returns data (including match) and match object.
        """
        with self.reading:
            while True:
                data = self.read_buffer.slice()
                match = regex.search(data)
                if match:
                    break
                self.read_buffer.enqueue((yield self.base.read(self.bufsize)))
            do_return((self.read_buffer.dequeue(match.end()), match))

    @async
    def write(self, data):
        """Write data

        Write data without blocking if write buffer's length less then doubled
        buffer size limit, if buffer's length is more then buffer size limit
        flush is started in the background.
        """
        with self.writing:  # state check
            self.write_buffer.enqueue(data)
        if len(self.write_buffer) > 2 * self.bufsize:
            yield self.flush()
        elif len(self.write_buffer) > self.bufsize:
            self.flush()()
        do_return(len(data))

    def write_schedule(self, data):
        """Enqueue data to write buffer

        Just enqueues data to write buffer (buffer's size limit would not be
        checked), flush need to be called manually.
        """
        self.write_buffer.enqueue(data)
        return len(data)

    @async
    def read_bytes(self):
        """Read bytes object
        """
        do_return((yield self.read_until_size(self.size_struct.unpack
                 ((yield self.read_until_size(self.size_struct.size)))[0])))

    def write_bytes(self, bytes):
        """Write bytes object to buffer
        """
        self.write_schedule(self.size_struct.pack(len(bytes)))
        self.write_schedule(bytes)

    @async
    def read_struct_list(self, struct, complex=None):
        """Read list of structures
        """
        struct_data = (yield self.read_until_size(self.size_struct.unpack((
                       yield self.read_until_size(self.size_struct.size)))[0]))
        if complex:
            do_return([struct.unpack(struct_data[offset:offset + struct.size])
                      for offset in range(0, len(struct_data), struct.size)])
        else:
            do_return([struct.unpack(struct_data[offset:offset + struct.size])[0]
                      for offset in range(0, len(struct_data), struct.size)])

    def write_struct_list(self, struct_list, struct, complex=None):
        """Write list of structures to buffer
        """
        self.write_schedule(self.size_struct.pack(len(struct_list) * struct.size))
        if complex:
            for struct_target in struct_list:
                self.write_schedule(struct.pack(*struct_target))
        else:
            for struct_target in struct_list:
                self.write_schedule(struct.pack(struct_target))

    @async
    def read_bytes_list(self):
        """Read array of bytes
        """
        bytes_list = []
        for size in (yield self.read_struct_list(self.size_struct, False)):
            bytes_list.append((yield self.read_until_size(size)))
        do_return(bytes_list)

    def write_bytes_list(self, bytes_list):
        """Write bytes array object to buffer
        """
        self.write_struct_list([len(bytes) for bytes in bytes_list],
                               self.size_struct, False)
        for bytes in bytes_list:
            self.write_schedule(bytes)


class Buffer(object):
    """Bytes FIFO buffer
    """
    def __init__(self):
        self.offset = 0
        self.chunks = deque()
        self.chunks_size = 0

    def slice(self, size=None, offset=None):
        """Get bytes with ``offset`` and ``size``
        """
        offset = offset or 0
        size = size or len(self)

        data = []
        data_size = 0

        # dequeue chunks
        size += self.offset + offset
        while self.chunks:
            if data_size >= size:
                break
            chunk = self.chunks.popleft()
            data.append(chunk)
            data_size += len(chunk)

        # re-queue merged chunk
        data = b''.join(data)
        self.chunks.appendleft(data)

        return data[self.offset + offset:size]

    def enqueue(self, data):
        """Enqueue "data" to buffer
        """
        if data:
            self.chunks.append(data)
            self.chunks_size += len(data)

    def dequeue(self, size=None, returns=None):
        """Dequeue "size" bytes from buffer

        Returns dequeued data if returns if True (or not set) otherwise None.
        """
        size = size or len(self)
        if not self.chunks:
            return b''

        data = []
        data_size = 0

        # dequeue chunks
        size = min(size + self.offset, self.chunks_size)
        while self.chunks:
            if data_size >= size:
                break
            chunk = self.chunks.popleft()
            data.append(chunk)
            data_size += len(chunk)

        if data_size == size:
            # no chunk re-queue
            self.chunks_size -= data_size
            offset, self.offset = self.offset, 0
        else:
            # If offset is beyond the middle of the chunk it will be split
            offset = len(chunk) - (data_size - size)
            if offset << 1 > len(chunk):
                chunk = chunk[offset:]
                offset, self.offset = self.offset, 0
            else:
                offset, self.offset = self.offset, offset

            # re-queue chunk
            self.chunks.appendleft(chunk)
            self.chunks_size += len(chunk) - data_size

        if returns is None or returns:
            return b''.join(data)[offset:size]

    def __len__(self):
        return self.chunks_size - self.offset

    def __bool__(self):
        return bool(self.chunks)
    __nonzero__ = __bool__

    def __str__(self):
        return 'Buffer(len:{})'.format(len(self))

    def __repr__(self):
        return str(self)



"""Haskell style do blocks from greenlets
"""
from functools import wraps
from .result import Result
try:
    import greenlet
except ImportError:
    greenlet = None

__all__ = ('do_green', 'bind_green',)


if greenlet is None:
    def do_green(Monad):
        def do(block):
            @wraps(block)
            def do_block(*args, **kw):
                raise NotImplementedError('greenlet module was not found')
            return do_block
        return do

    def bind_green(monad):
        raise NotImplementedError('greenlet module was not found')

else:
    def do_green(Monad):
        """Greenlet based do block

        Create Haskell style do blocks from greenlet coroutine. One obvious
        limitation of this implementation it cannot be used with monad which call
        bound function multiple times inside bind implementation (e.g. List monad).
        """
        unit = Monad.unit

        def do(block):
            @wraps(block)
            def do_block(*args, **kw):
                def do_next(result):
                    coro.parent = greenlet.getcurrent()
                    val, err = (result.pair if isinstance(result, Result) else
                               (result, None))
                    try:
                        result = (coro.switch(val) if err is None else
                                  coro.throw(*err))
                        if coro.dead:
                            return unit(Result.from_value(result))
                        else:
                            return result.__monad__().bind(do_next)
                    except Exception:
                        return unit(Result.from_current_error())
                coro = _do_greenlet(lambda _: block(*args, **kw))
                return do_next(None)
            return do_block
        return do

    class _do_greenlet(greenlet.greenlet):
        """Do specific greenlet
        """

    def bind_green(monad):
        """Bind monad inside greenlet do block
        """
        curr = greenlet.getcurrent()
        if not isinstance(curr, _do_greenlet) or curr.parent is None:
            raise RuntimeError('bind outside of do_greenlet')
        return curr.parent.switch(monad)



import errno
import select
from .. import PRETZEL_POLLER

__all__ = ('Poller', 'POLL_READ', 'POLL_WRITE', 'POLL_URGENT', 'POLL_DISCONNECT', 'POLL_ERROR',)

EPOLLIN = 0x001
EPOLLPRI = 0x002
EPOLLOUT = 0x004
EPOLLRDNORM = 0x040
EPOLLRDBAND = 0x080
EPOLLWRNORM = 0x100
EPOLLWRBAND = 0x200
EPOLLMSG = 0x400
EPOLLERR = 0x008
EPOLLHUP = 0x010
EPOLLRDHUP = 0x2000
EPOLLONESHOT = 1 << 30
EPOLLET = 1 << 31

POLL_READ = EPOLLIN
POLL_WRITE = EPOLLOUT
POLL_URGENT = EPOLLPRI
POLL_ERROR = EPOLLERR | EPOLLHUP
POLL_DISCONNECT = EPOLLHUP


class Poller (object):
    @classmethod
    def from_name(cls, name=None):
        name = name or PRETZEL_POLLER

        if name == 'epoll' and hasattr(select, 'epoll'):
            return EPollPoller()
        elif name == 'kqueue' and hasattr(select, 'kqueue'):
            return KQueuePoller()
        elif name == 'select':
            return SelectPoller()

        raise NotImplementedError('poller method is not support: {}'.format(name))

    def register(self, fd, mask):
        raise NotImplementedError()

    def modify(self, fd, mask):
        raise NotImplementedError()

    def unregister(self, fd):
        raise NotImplementedError()

    def poll(self, timeout):
        raise NotImplementedError()

    def dispose(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return '{}()'.format(type(self).__name__)

    def __repr__(self):
        return str(self)


class EPollPoller (Poller):
    """Poller base on epoll
    """
    def __init__(self):
        self.fds = {}
        self.epoll = select.epoll()

        from ..stream.file import fd_close_on_exec
        fd_close_on_exec(self.epoll.fileno(), True)

    def register(self, fd, mask):
        self.epoll.register(fd, mask)
        self.fds.setdefault(fd, True)

    def modify(self, fd, mask):
        self.epoll.modify(fd, mask)

    def unregister(self, fd):
        if self.fds.pop(fd, None):
            self.epoll.unregister(fd)

    def poll(self, timeout):
        if not self.fds and timeout < 0:
            raise StopIteration()  # would have blocked indefinitely

        try:
            return self.epoll.poll(timeout)
        except (IOError, OSError) as error:
            if error.errno == errno.EINTR:
                return tuple()
            raise

    def dispose(self):
        self.epoll.close()

    def __str__(self):
        return '{}(fds:{}, fd:{})'.format(type(self).__name__, len(self.fds),
                                          self.epoll.fileno())


class SelectPoller(Poller):
    """Poller base on select
    """
    SUPPORTED_MASK = POLL_READ | POLL_WRITE

    def __init__(self):
        self.read = set()
        self.write = set()
        self.error = set()

    def register(self, fd, mask):
        if mask | self.SUPPORTED_MASK != self.SUPPORTED_MASK:
            raise ValueError('unsupported event mask: {}'.format(mask))

        self.error.add(fd)
        if mask & POLL_READ:
            self.read.add(fd)
        if mask & POLL_WRITE:
            self.write.add(fd)

    def modify(self, fd, mask):
        self.unregister(fd)
        self.register(fd, mask)

    def unregister(self, fd):
        self.read.discard(fd)
        self.write.discard(fd)
        self.error.discard(fd)

    def poll(self, timeout):
        if not self.error and timeout < 0:
            raise StopIteration()  # would have blocked indefinitely

        try:
            read, write, error = select.select(self.read, self.write, self.error,
                                               timeout if timeout >= 0 else None)
        except (OSError, IOError) as error:
            if error.errno == errno.EINTR:
                return tuple()
            raise
        except select.error as error:
            if error.args[0] == errno.EINTR:
                return tuple()
            raise

        events = {}
        for fd in read:
            events[fd] = events.get(fd, 0) | POLL_READ
        for fd in write:
            events[fd] = events.get(fd, 0) | POLL_WRITE
        for fd in error:
            events[fd] = events.get(fd, 0) | POLL_ERROR

        return events.items()

    def __str__(self):
        return '{}(read:{}, write:{})'.format(type(self).__name__,
                                              len(self.read), len(self.write))


class KQueuePoller(SelectPoller):
    pass



"""Bootstrap module

Bootstrap package hierarchy from single string.
"""
import os
import io
import re
import sys
import zlib
import json
import types
import struct
import pickle
import inspect
import binascii
import textwrap
import importlib

__all__ = ('BootImporter', 'BootLoader', 'boot_module', 'boot_boot',
           'boot_binary', 'boot_pack')


class BootImporter(object):
    """Bootstrap importer
    """
    __tag__ = '1f43bbd9-36e0-4084-84e5-b7fb14fdb1bd'

    def __init__(self, loaders=None):
        self.loaders = loaders or {}

    @classmethod
    def from_stream(cls, stream):
        loaders = {}
        for _ in range(struct.unpack('>L', stream.read(struct.calcsize('>L')))[0]):
            loader = BootLoader.from_stream(stream)
            loaders[loader.name] = loader
        return cls(loaders)

    def to_stream(self, stream):
        stream.write(struct.pack('>L', len(self.loaders)))
        for loader in self.loaders.values():
            loader.to_stream(stream)
        return stream

    @classmethod
    def from_bytes(cls, bytes):
        return cls.from_stream(io.BytesIO(bytes))

    def to_bytes(self):
        stream = io.BytesIO()
        self.to_stream(stream)
        return stream.getvalue()

    @classmethod
    def from_modules(cls, modules=None):
        instance = cls()
        if modules:
            for module in modules:
                instance.add_module(module)
        else:
            instance.add_module(__package__ or __name__.partition('.')[0])
        return instance

    def add_module(self, module):
        """Add module or package

        Only capable to add modules witch reside on disk as plain python files
        or in tomb.
        """
        if isinstance(module, str):
            # convert module name to module
            module = sys.modules.get(module, None) or importlib.import_module(module)

        # find top level package
        modname = (getattr(module, '__package__', '') or module.__name__).partition('.')[0]
        if modname != module.__name__:
            module = importlib.import_module(modname)

        if modname in self.loaders:
            return  # skip already imported packages

        loader = getattr(module, '__loader__', None)
        for importer in sys.meta_path:
            if modname in getattr(importer, 'loaders', {}):
                # package was loaded with BootImporter
                for name, loader in importer.loaders.items():
                    if name.startswith(modname):
                        self.loaders[name] = loader
                return

        # find package file
        filename = inspect.getsourcefile(module)
        if not filename:
            raise ValueError('module does not have sources: {}'.format(modname))

        # Use file name to determine if it is a package instead of loader.is_package
        # because is_package incorrectly handles __main__ module.
        if os.path.basename(filename).lower() == '__init__.py':
            root = os.path.dirname(filename)
            for path, dirs, files in os.walk(root):
                for file in files:
                    if not file.lower().endswith('.py'):
                        continue
                    filename = os.path.join(path, file)
                    source = self.read_source(filename)
                    name = (modname if os.path.samefile(path, root) else
                            '.'.join((modname, os.path.relpath(path, root).replace('/', '.'))))
                    if file.lower() == '__init__.py':
                        self.add_source(name, source, filename, True)
                    else:
                        self.add_source('.'.join((name, file[:-3])), source, filename, False)
        else:
            self.add_source(modname, self.read_source(filename), filename, False)

    def add_source(self, name, source, filename, ispkg=False, pkg=None):
        """Add source file as specified module
        """
        self.loaders[name] = BootLoader(name, source, filename, ispkg)

    def find_module(self, name, path=None):
        return self.loaders.get(name, None)

    def bootstrap(self, init=None, *init_a, **init_kw):
        """Create bootstrap source for this importer

        Initialization function ``init`` and its arguments ``init_a``,
        ``init_kw`` must be pickle-able objects and its required modules must be
        added to the importer.
        """
        if init and inspect.getmodule(init).__name__ not in self.loaders:
            raise ValueError('initialization function must reside in added modules')

        source = StringIO()
        source.write(textwrap.dedent("""\
            {}
            _boot.BootImporter.from_bytes({}).install()
            """).format(boot_boot('_boot'), boot_binary(self.to_bytes())))
        if init is not None:
            source.write(textwrap.dedent("""
                import pickle
                init, init_a, init_kw = pickle.loads({})
                if init is not None:
                    init(*init_a, **init_kw)
                """).format(boot_binary(pickle.dumps((init, init_a, init_kw)))))
        return source.getvalue()

    def install(self):
        """Install importer into system path
        """
        if self not in sys.meta_path:
            sys.meta_path.insert(0, self)

    def dispose(self):
        if self in sys.meta_path:
            self.meta_path.remove(self)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    @staticmethod
    def read_source(filename):
        """Read source from file name
        """
        encoding = 'utf-8'
        encoding_pattern = re.compile(b'coding[:=]\s*([-\w.]+)')  # PEP: 0263

        source = io.BytesIO()
        with io.open(filename, 'rb') as stream:
            for line in stream:
                if line.startswith(b'#'):
                    match = encoding_pattern.search(line)
                    if match:
                        encoding = match.group(1).decode()
                        source.write(b'\n')
                        continue
                source.write(line)
        if PY2:
            return source.getvalue()
        else:
            return source.getvalue().decode(encoding)


class BootLoader(object):
    __slots__ = ('name', 'source', 'filename', 'ispkg', 'pkg',)

    def __init__(self, name, source, filename, ispkg, pkg=None):
        self.name = name
        self.source = source
        self.filename = filename
        self.ispkg = ispkg
        self.pkg = pkg or name if ispkg else name.rpartition('.')[0]

    @classmethod
    def from_stream(cls, stream):
        data = json.loads(zlib.decompress(stream.read(struct.unpack('>L',
                          stream.read(struct.calcsize('>L')))[0])).decode('utf-8'))
        if PY2:
            # Convert unicode string to bytes, avoids problems with traceback
            # creation and error reporting.
            return cls(data['name'].encode('utf-8'),
                       data['source'].encode('utf-8'),
                       data['filename'].encode('utf-8'),
                       data['ispkg'],
                       data['pkg'].encode('utf-8'))
        else:
            return cls(data['name'], data['source'], data['filename'],
                       data['ispkg'], data['pkg'])

    def to_stream(self, stream):
        data = zlib.compress(json.dumps({
            'name': self.name,
            'source': self.source,
            'filename': self.filename,
            'ispkg': self.ispkg,
            'pkg': self.pkg,
        }).encode('utf-8'), 9)
        stream.write(struct.pack('>L', len(data)))
        stream.write(data)
        return stream

    def __call__(self):
        return self.load_module(self.name)

    def load_module(self, name):
        module = sys.modules.get(name)
        if module is not None:
            return module
        if name != self.name:
            raise ImportError('loader cannot handle {}'.format(name))

        module = types.ModuleType(self.name)
        module.__package__ = self.pkg
        module.__file__ = self.filename
        module.__loader__ = self
        if self.ispkg:
            module.__path__ = []

        module.__initializing__ = True
        sys.modules[name] = module
        try:
            execute(compile(self.source, module.__file__, 'exec'), module.__dict__)
            return module
        except Exception:
            sys.modules.pop(name, None)
            raise
        finally:
            module.__initializing__ = False

    def is_package(self, name):
        if name != self.name:
            raise ImportError('loader cannot handle {}'.format(name))
        return self.ispkg

    def get_source(self, name):
        if name != self.name:
            raise ImportError('loader cannot handle {}'.format(name))
        return self.source

    def get_code(self, name):
        if name != self.name:
            raise ImportError('loader cannot handle {}'.format(name))
        return compile(self.source, self.filename, 'exec')

    def __str__(self):
        return '{}(name:{}, package:{})'.format(type(self).__name__,
                                                self.name, self.pkg)

    def __repr__(self):
        return str(self)


def boot_module(name, source, filename):
    """Bootstrap python source as module

    Returns python source, witch when executed allows to import specified
    ``source`` as module with specified ``name``.
    """
    source_payload = textwrap.dedent("""\
        import sys
        import zlib
        import types
        import binascii
        {sep}
        def load():
            module = types.ModuleType("{name}")
            module.__file__ = "{filename}"
            module.__package__ = "{name}"
            sys.modules["{name}"] = module
            try:
                code = compile({source}, module.__file__, "exec")
                if sys.version_info[0] > 2:
                    exec(code, module.__dict__)
                else:
                    exec("exec code in module.__dict__")
                return module
            except Exception:
                sys.modules.pop("{name}")
                raise
        try:
            {name} = load()
        finally:
            del load
        """)
    return source_payload.format(name=name, filename=filename, sep='\n',
                                 source=boot_binary(source.encode('utf-8')))


def boot_pack(source, encoding=None):
    """Pack python source

    Returns packed source which can be passed to python interpreter via '-c'
    option and be executed. Resulted source does not contain only spaces and
    ASCII symbols.
    """
    encoding = encoding or 'utf-8'
    source_fmt = 'import zlib,binascii;exec(zlib.decompress(binascii.a2b_base64(b"{}")))'
    source_packed = binascii.b2a_base64(zlib.compress(source.encode(encoding), 9)).strip()
    return source_fmt.format(source_packed.decode())


def boot_binary(data):
    """Binary data trampoline

    Returns source code which unpack into provided binary ``data``.
    """
    return ('zlib.decompress(binascii.a2b_base64(b"\\\n{}"))'.format(
            '\\\n'.join(textwrap.wrap(binascii.b2a_base64(zlib.compress(data, 9))
            .strip().decode('utf-8'), 78))))


def boot_boot(name):
    """Bootstrap this module

    Returns python source, witch when executed allows to import this module
    by specified "name".
    """
    module = sys.modules[__name__]
    return boot_module(name, inspect.getsource(module),
                       inspect.getsourcefile(module))


#  python version compatibility
if sys.version_info[0] > 2:
    import builtins
    execute = getattr(builtins, "exec")
    del builtins

    def reraise(tp, value, tb=None):
        """Re-raise exception
        """
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    StringIO = io.StringIO
    PY2 = False

else:
    def execute(code, globs=None, locs=None):
        """Execute code in a name space.
        """
        if globs is None:
            frame = sys._getframe(1)
            globs = frame.f_globals
            if locs is None:
                locs = frame.f_locals
            del frame
        elif locs is None:
            locs = globs
        exec("""exec code in globs, locs""")

    exec("""def reraise(tp, value, tb=None):
        raise tp, value, tb""")

    StringIO = io.BytesIO
    PY2 = True


def main():  # pragma: no cover
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--main', dest='main', default='',
                        help='file used as __main__ module')
    parser.add_argument('modules', nargs=argparse.REMAINDER, help='modules')
    opts = parser.parse_args()

    sys.stdout.write('#! /usr/bin/env python\n' if opts.main else '')
    sys.stdout.write(BootImporter.from_modules(opts.modules or None).bootstrap())
    sys.stdout.write('\n')

    if opts.main:
        sys.stdout.write(BootImporter.read_source(opts.main))

if __name__ == '__main__':
    main()



"""Remoting benchmarks
"""
import itertools
from .conn import ForkConnection
from .proxy import proxify
from ..monad import async, async_all
from ..bench import Benchmark


class FuncBench(Benchmark):
    """Benchmark function call
    """
    def __init__(self):
        Benchmark.__init__(self, 'remoting.func', 1)
        self.conn = None

    @async
    def init(self):
        self.conn = yield ForkConnection()
        self.func = self.conn(remote)()
        if ((yield self.func), (yield self.func)) != (1, 2):
            raise ValueError('initialization test failed')

    def body(self):
        return self.func

    def dispose(self):
        conn, self.conn = self.conn, None
        if conn:
            conn.dispose()


class FuncAsyncBench(FuncBench):
    """Benchmark asynchronous function call
    """
    def __init__(self):
        Benchmark.__init__(self, 'remoting.func_async', 512)
        self.conn = None

    def body(self):
        return async_all((self.func,) * self.factor)


class MethodBench (Benchmark):
    """Benchmark proxy method call
    """
    def __init__(self):
        Benchmark.__init__(self, 'remoting.method', 1)
        self.conn = None

    @async
    def init(self):
        self.conn = yield ForkConnection()
        self.proxy = yield proxify(self.conn(Remote)())
        self.method = self.proxy.method()
        if ((yield self.method), (yield self.method)) != (1, 2):
            raise ValueError('initialization test failed')

    def body(self):
        return self.method

    def dispose(self):
        proxy, self.proxy = self.proxy, None
        if proxy:
            proxy.dispose()
        conn, self.conn = self.conn, None
        if conn:
            conn.dispose()


class MethodAsyncBench(MethodBench):
    """Benchmark asynchronous proxy method call
    """
    def __init__(self):
        Benchmark.__init__(self, 'remoting.method_async', 1024)
        self.conn = None

    def body(self):
        return async_all((self.method,) * self.factor)


def remote():
    return next(fn_count)
fn_count = itertools.count(1)


class Remote(object):
    def __init__(self):
        self.count = itertools.count(1)

    def method(self):
        return next(self.count)


def load_bench(runner):
    """Load benchmarks
    """
    for bench in ((FuncBench(), FuncAsyncBench(),
                   MethodBench(), MethodAsyncBench(),)):
        runner.add(bench)



import math
import unittest
import itertools
from heapq import heappush, heappop
from ..do import do_return
from ..async import async, async_block, async_all, async_any, async_limit
from ..result import Result
from ...event import Event

__all__ = ('ContTest',)


class ContTest(unittest.TestCase):
    """Continuation monad unit tests
    """
    def test_normal(self):
        rets, rets_ref = [], []
        e0, e1 = Event(), Event()

        @async
        def test_async(result):
            rets.append((yield e0))
            rets.append('0-1')
            rets.append((yield e1))
            rets.append('1-2')
            rets.append((yield e0))
            rets.append('2-R')
            do_return(result)

        test_async('done')(lambda val: rets.append(val))
        self.assertEqual(rets, rets_ref)

        e1(1)
        self.assertFalse(rets, rets_ref)
        e0(2)
        rets_ref.extend([2, '0-1'])
        self.assertEqual(rets, rets_ref)
        e1(Result.from_value(3))
        rets_ref.extend([3, '1-2'])
        self.assertEqual(rets, rets_ref)
        e0(4)
        rets_ref.extend([4, '2-R', Result.from_value('done')])
        self.assertEqual(rets, rets_ref)

    def test_error(self):
        rets, rets_ref = [], []
        ev = Event()
        error = ValueError('test')

        @async
        def test_async():
            try:
                rets.append((yield ev))
            except Exception as error:
                rets.append(error)
            rets.append((yield ev))
            rets.append((yield ev))

        test_async()(lambda val: rets.extend(('done', val,)))
        self.assertEqual(rets, rets_ref)

        ev(Result.from_exception(error))
        rets_ref.append(error)
        self.assertEqual(rets, rets_ref)

        ev('value')
        rets_ref.append('value')
        self.assertEqual(rets, rets_ref)

        ev(Result.from_exception(error))
        rets_ref.extend(('done', Result.from_exception(error),))
        self.assertEqual(rets, rets_ref)

    def test_block(self):
        rets, rets_ref = [], []
        hs = []
        exc = ValueError('test')

        def test_async(exc=None):
            @async_block
            def cont(ret):
                if exc is None:
                    hs.append(ret)
                else:
                    raise exc
            return cont

        test_async(exc)(lambda val: rets.append(val))
        rets_ref.append(Result.from_exception(exc))
        self.assertEqual(rets, rets_ref)

        test_async()(lambda val: rets.append(val))
        hs[-1]('done')
        rets_ref.append(Result.from_value('done'))
        self.assertEqual(rets, rets_ref)

        test_async()(lambda val: rets.append(val))
        hs[-1](Result.from_value('done value'))
        rets_ref.append(Result.from_value('done value'))
        self.assertEqual(rets, rets_ref)

    def test_all(self):
        rets = []
        e0, e1 = Event(), Event()

        async_all((e0, e1))(lambda val: rets.append(val))
        self.assertEqual(len(e0.handlers), 1)
        self.assertEqual(len(e1.handlers), 1)

        e1('two')
        e0('one')
        self.assertFalse(e0.handlers)
        self.assertFalse(e1.handlers)
        self.assertEqual(rets.pop().value, ('one', 'two'))

        async_all(tuple())(lambda val: rets.append(val))
        self.assertEqual(rets.pop().value, tuple())

    def test_any(self):
        rets = []
        e0, e1 = Event(), Event()

        async_any((e0, e1))(lambda val: rets.append(val))
        self.assertEqual(len(e0.handlers), 1)
        self.assertEqual(len(e1.handlers), 1)

        e0('done')
        self.assertFalse(e0.handlers)
        self.assertTrue(e1.handlers)
        self.assertEqual(rets.pop().value, 'done')

        e1('done_next')
        self.assertFalse(e1.handlers)

        with self.assertRaises(ValueError):
            async_any(tuple())

    def test_limit(self):
        timer = Timer()
        timer_limit_10 = async_limit(10)(lambda val: timer(val))

        count = 1024
        res = async_all(timer_limit_10(1) for i in range(count)).future()
        for i in range(count):
            if res.completed:
                res.value
                break
            timer.tick()
        self.assertEqual(timer.time, math.ceil(count / 10.))


class Timer(object):
    def __init__(self):
        self.time = 0
        self.uid = itertools.count()
        self.queue = []

    def __call__(self, time):
        @async_block
        def cont(ret):
            heappush(self.queue, (self.time + time, next(self.uid), ret))
        return cont

    def tick(self):
        self.time += 1
        while self.queue:
            time, _, ret = self.queue[0]
            if time > self.time:
                return
            else:
                heappop(self.queue)
                ret(self.time)



"""Shell connection

Connection with process over standard output and input, error stream
is untouched.
"""
import io
import os
import sys
import socket
import pickle
import textwrap

from .stream import StreamConnection
from ..importer import Importer
from ... import PRETZEL_BUFSIZE, PRETZEL_RECLIMIT
from ...monad import async
from ...process import Process, PIPE
from ...boot import BootImporter, boot_pack, __name__ as boot_name
from ...core import Core
from ...stream import BufferedFile

__all__ = ('ShellConnection',)


class ShellConnection(StreamConnection):
    """Shell connection

    Connection with process over standard output and input, error stream
    is untouched.
    """
    def __init__(self, command=None, escape=None, py_exec=None,
                 environ=None, bufsize=None, hub=None, core=None):
        StreamConnection.__init__(self, hub=hub, core=core)

        self.bufsize = bufsize
        self.py_exec = py_exec or sys.executable
        self.command = command or []
        self.command.extend((self.py_exec, '-c', '\'{}\''.format(shell_tramp)
                            if escape else shell_tramp))
        self.environ = environ
        self.process = None

    @async
    def do_connect(self, target):
        """Fork connect implementation

        Target is pickle-able and call-able which will be called upon successful
        connection with connection this as its only argument.
        """
        def preexec():  # pragma: no cover
            os.chdir('/')
            os.setsid()
        self.process = yield (self.dispose.add(Process(self.command, stdin=PIPE,
                                               stdout=PIPE, preexec=preexec,
                                               bufsize=self.bufsize, core=self.core)))

        # check remote python version
        version = yield self.process.stdout.read_bytes()
        if version.decode() != str(sys.version_info[0]):
            raise RuntimeError('remote python major version mismatch local:{} remote:{}'
                               .format(sys.version_info[0], version.decode()))

        # send environment data
        environ = {
            'PRETZEL_RECLIMIT': str(PRETZEL_RECLIMIT),
            'PRETZEL_BUFSIZE': str(PRETZEL_BUFSIZE),
        }
        if self.environ:
            environ.update(self.environ)
        self.process.stdin.write_bytes(pickle.dumps(environ))

        # send boot data
        boot_data = (BootImporter.from_modules().bootstrap(
                     shell_conn_init, self.bufsize).encode('utf-8'))
        self.process.stdin.write_bytes(boot_data)
        yield self.process.stdin.flush()

        yield StreamConnection.do_connect(self, (self.process.stdout,
                                                 self.process.stdin))
        yield self(os.chdir)('/')

        # install importer
        self.dispose.add((yield Importer.create_remote(self)))
        self.module_map['_boot'] = boot_name

        # update flags
        self.flags['pid'] = yield self(os.getpid)()
        self.flags['host'] = yield self(socket.gethostname)()


def shell_conn_init(bufsize):  # pragma: no cover
    """Shell connection initialization function
    """
    # Make sure standard output and input won't be used. As it is now used
    # for communication.
    in_fd = os.dup(0)
    with io.open(os.devnull, 'r') as stdin:
        os.dup2(stdin.fileno(), 0)
    out_fd = os.dup(1)
    os.dup2(sys.stderr.fileno(), 1)

    with Core.local() as core:
        # initialize connection
        conn = StreamConnection(core=core)
        conn.flags['pid'] = os.getpid()
        conn.flags['host'] = socket.gethostname()
        conn.dispose.add_action(lambda: core.schedule()(lambda _: core.dispose()))

        # connect
        in_stream = BufferedFile(in_fd, bufsize=bufsize, core=core)
        in_stream.close_on_exec(True)
        out_stream = BufferedFile(out_fd, bufsize=bufsize, core=core)
        out_stream.close_on_exec(True)
        conn.connect((in_stream, out_stream))()

        if not core.disposed:
            core()

shell_tramp = boot_pack(textwrap.dedent("""\
    import os, io, sys, struct, pickle
    size_struct = struct.Struct("{size_format}")
    # send version
    version = str(sys.version_info[0]).encode()
    os.write(1, size_struct.pack(len(version)))
    os.write(1, version)
    # receiver payload
    with io.open(0, "rb", buffering=0, closefd=False) as stream:
        def read_bytes():
            size_data = stream.read(size_struct.size)
            if len(size_data) < size_struct.size:
                sys.exit(0)
            size = size_struct.unpack(size_data)[0]
            data = io.BytesIO()
            while size > data.tell():
                chunk = stream.read(size - data.tell())
                if not chunk:
                    raise ValueError("payload is incomplete")
                data.write(chunk)
            return data.getvalue()
        env_data = read_bytes()
        boot_data = read_bytes()
    os.environ.update(pickle.loads(env_data))
    exec(boot_data.decode("utf-8"))
    """.format(size_format=BufferedFile.size_struct.format.decode())))



"""Connection mesh

Create composite connection and work with it the same way as with single
connection.
"""
import math
import random
import functools
from .fork import ForkConnection
from .ssh import SSHConnection
from ...monad import List, Proxy, async, async_all, do_return

__all__ = ('composite_conn', 'composite_ssh_conn', 'composite_fork_conn',)


@async
def composite_conn(conn_facts, mesh=None):
    """Create composite connection from list of connection factories.

    Arguments:
        conn_facts: List of connection factories. Keep in mind that this factories
                    must be pickle-able in case of non flat mesh (you can use
                    functools.partial to keep connection type pickle-able)
        mesh: Mesh of composite connection. Available options are
             'flat' - direct connections (is default)
             'tree:factor' - mesh has a form of tree with factor children.
    Returns:
        Continuation list of connections
    """
    mesh = mesh or 'flat'
    conn_facts = tuple(conn_facts)

    if mesh == 'flat':
        conns = yield ContList(conn_facts)()
        do_return((yield ContList(conn(conn) for conn in conns)))  # to proxy

    elif mesh.startswith('tree:'):
        tree_factor = int(mesh.partition(':')[2])
        if tree_factor < 1:
            raise ValueError('tree factor must be positive')
        tree = Tree.from_list(conn_facts, tree_factor)

        @async
        def connect(conn_fact, parent):
            if conn_fact is None:
                do_return(None)
            if parent is None:
                conn = yield conn_fact()
                conn = yield conn(conn)  # to proxy
            else:
                conn = yield ~parent(conn_fact)()
            do_return(conn)
        conns = yield tree(connect, None)

        do_return(ContList(conns[1:]))  # strip None from beginning


def composite_ssh_conn(hosts, mesh=None, **conn_opts):
    """Create composite ssh connection from list of hosts

    Arguments:
        hosts: List of hosts
        mesh: Mesh of composite connection (see composite_conn)
        **conn_opts: Common connection options
    """
    if isinstance(hosts, str):
        hosts = (hosts,)
    conn_facts = (functools.partial(SSHConnection, host, **conn_opts)
                  for host in hosts)
    return composite_conn(conn_facts, mesh=mesh)


def composite_fork_conn(count, mesh=None, **conn_opts):
    """Create `count` composite fork connection

    Arguments:
        count: Number of fork connections
        mesh: Mesh of composite connection (see composite_conn)
        **conn_opts: Common connection options
    """
    conn_facts = (functools.partial(ForkConnection, **conn_opts),) * count
    return composite_conn(conn_facts, mesh=mesh)


class ContList(Proxy):
    """List of continuation monads

    Behaves as proxy on list monad but when used as monad assumes that
    it contains list of continuation which will be executed simultaneously when
    executed.
    """
    __slots__ = Proxy.__slots__

    def __init__(self, items):
        Proxy.__init__(self, List(*items))

    def __monad__(self):
        return (async_all(Proxy.__monad__(self)).map_val
                         (lambda res: res.map_val(ContList)))

    def __iter__(self):
        """Iterate over containing items.
        """
        return iter(Proxy.__monad__(self))


class Tree(object):
    """Helper three structure
    """
    __slots__ = ('value', 'children',)

    def __init__(self, value, children):
        self.value = value
        self.children = children

    @classmethod
    def from_list(cls, items, factor, root=None):
        """Build tree

        Creates tree of items with factor children.
        """
        def build(vals):
            val, vals = vals[0], vals[1:]
            if vals:
                bucket = int(math.ceil(len(vals) / float(factor)))
                return cls(val, tuple(build(vals[bucket * i:bucket * (i + 1)])
                           for i in range(factor) if bucket * i < len(vals)))
            else:
                return cls(val, tuple())
        items = list(items)
        random.shuffle(items)
        return build([root] + items)

    @async
    def __call__(self, func, val):
        """Apply asynchronous function and collect results

        Returns list of results of functions application on each node.
        """
        v = yield func(self.value, val)
        if self.children:
            css = yield async_all(child(func, v) for child in self.children)
            do_return([v] + [c for cs in css for c in cs])
        else:
            do_return([v])

    def __str__(self):
        return '{{{}: {}}}'.format(self.value, list(self.children))

    def __repr__(self):
        return str(self)



import unittest
from ..list import List

__all__ = ('MonadTest',)


class MonadTest(unittest.TestCase):
    def test_sequence(self):
        # sequence :: (Monad m) => [m a] -> m [a]
        self.assertEqual(List.sequence([List(1, 2), List(-1, -2)]),
                         List((1, -1), (1, -2), (2, -1), (2, -2)))

    def test_map(self):
        # map :: (Monad m) => (a -> m a) -> [a] -> m [a]
        self.assertEqual(List.map(lambda a: List(-a, a), [1, 2]),
                         List((-1, -2), (-1, 2), (1, -2), (1, 2)))

    def test_filter(self):
        # filter :: (Monad m) => (a -> m Bool) -> [a] -> m [a]
        self.assertEqual(List.filter(lambda a: List(a % 2 == 0), range(10)),
                         List((0, 2, 4, 6, 8),))

    def test_fold(self):
        # fold :: (Monad m) => (a -> b -> m a) -> a -> [b] -> m a
        self.assertEqual(List.fold(lambda a, b: List(a + b), 0, range(10)),
                         List(45))

    def test_lift_func(self):
        # lift_func :: (Monad m) => (a0 -> ... -> r) -> (m a0 -> ... -> m r)
        list_add = List.lift_func(lambda a, b: a + b)
        self.assertEqual(list_add(List(0, 1), List(0, 1)), List(0, 1, 1, 2))

    def test_ap(self):
        # ap :: (Monad a) => m (a -> b) -> (m a -> m b)
        self.assertEqual(List.ap(List(lambda a: a * 2))(List(1, 2)), List(2, 4))



import os
import time
import unittest
from . import waitpid
from .core import FileQueue, TimeQueue, CORE_TIMEOUT
from .poll import POLL_READ, POLL_WRITE, POLL_URGENT, POLL_DISCONNECT, EPOLLERR
from ..uniform import BrokenPipeError, ConnectionError
from ..tests import async_test

__all__ = ('FileTest', 'TimeQueueTest', 'ProcQueueTest',)


class FileTest(unittest.TestCase):
    def test_dummy(self):
        polls, rets = [], []
        ret = lambda tag: lambda res: rets.append((tag, res))
        file = FileQueue('fd', self.DummyPoller(lambda *a: polls.append(a)))

        file.on(POLL_READ | POLL_WRITE)(ret('rw'))
        file.on(POLL_READ)(ret('r'))
        self.assertEqual(polls, [('reg', 'fd', POLL_READ | POLL_WRITE)])
        self.assertEqual(rets[0][0], 'r')
        with self.assertRaises(ValueError):  # intersecting mask
            rets[0][1].value
        del rets[:]

        file.on(POLL_URGENT)(ret('u'))
        self.assertEqual(len(polls), 2)
        self.assertEqual(polls[-1], ('mod', 'fd', POLL_READ | POLL_WRITE | POLL_URGENT))

        file(POLL_READ)
        self.assertEqual(rets[0][0], 'rw')
        self.assertEqual(rets[0][1].value, POLL_READ)
        self.assertEqual(len(polls), 3)
        self.assertEqual(polls[-1], ('mod', 'fd', POLL_URGENT))

        file(POLL_READ)
        self.assertEqual(len(rets), 1)

        file(POLL_URGENT)
        self.assertEqual(len(rets), 2)
        self.assertEqual(rets[-1][0], 'u')
        self.assertEqual(len(polls), 4)
        self.assertEqual(polls[-1], ('unreg', 'fd'))

        file.on(POLL_WRITE)(ret('w'))
        self.assertEqual(len(polls), 5)
        self.assertEqual(polls[-1], ('reg', 'fd', POLL_WRITE))

        file(POLL_DISCONNECT)
        self.assertEqual(len(polls), 6)
        self.assertEqual(polls[-1], ('unreg', 'fd'))
        self.assertEqual(len(rets), 3)
        self.assertEqual(rets[-1][0], 'w')
        with self.assertRaises(BrokenPipeError):
            rets[-1][1].value

        file.on(POLL_READ)(ret('r'))
        file.on(POLL_WRITE)(ret('w'))
        self.assertEqual(len(polls), 8)
        self.assertEqual(polls[-2:], [('reg', 'fd', POLL_READ),
                                      ('mod', 'fd', POLL_READ | POLL_WRITE)])

        file(EPOLLERR)
        self.assertEqual(len(polls), 9)
        self.assertEqual(polls[-1], ('unreg', 'fd'))
        self.assertEqual(len(rets), 5)
        self.assertEqual(rets[-2][0], 'r')
        with self.assertRaises(ConnectionError):
            rets[-2][1].value
        self.assertEqual(rets[-1][0], 'w')
        with self.assertRaises(ConnectionError):
            rets[-1][1].value

    class DummyPoller(object):
        def __init__(self, hook):
            self.hook = hook

        def register(self, fd, mask):
            self.hook('reg', fd, mask)

        def modify(self, fd, mask):
            self.hook('mod', fd, mask)

        def unregister(self, fd):
            self.hook('unreg', fd)


class TimeQueueTest(unittest.TestCase):
    def test(self):
        def res():
            ret = tuple(rets)
            del rets[:]
            return ret
        ret = lambda tag: lambda res: rets.append((tag, res.value))
        rets = []

        timer = TimeQueue()
        timer.on(1)(ret('1'))
        timer.on(2)(ret('2'))

        timer(.5)
        self.assertEqual(len(res()), 0)
        self.assertEqual(timer.timeout(.5), .5)

        timer(1.5)
        self.assertEqual(res(), (('1', 1),))

        timer(2)
        self.assertEqual(res(), (('2', 2),))
        self.assertEqual(timer.timeout(2), CORE_TIMEOUT)

        timer.on(1)(ret('1:0'))
        timer.on(1)(ret('1:1'))
        timer.on(2)(ret('2'))
        self.assertEqual(timer.timeout(3), 0)

        timer(3)
        self.assertEqual(res(), (('1:0', 1), ('1:1', 1), ('2', 2)))


class ProcQueueTest(unittest.TestCase):
    @async_test
    def test(self):
        pid = os.fork()
        if pid:
            self.assertEqual((yield waitpid(pid)), pid & 0xff)
        else:
            time.sleep(.3)
            os.execvp('/bin/sh', ['/bin/sh', '-c', 'exit {}'.format(os.getpid() & 0xff)])



"""Benchmark class
"""
from __future__ import print_function
import sys
import time
from operator import add
from functools import reduce
from .monad import async, do_return
from .app import app

__all__ = ('Benchmark', 'TextBenchmarkRunner',)


class Benchmark(object):
    default_error = 0.01
    default_time = time.time
    min_count = 5
    max_time = 15

    def __init__(self, name=None, factor=None):
        self.name = name or type(self).__name__
        self.factor = factor or 1

    @async
    def init(self):
        """Initialize benchmark
        """

    @async
    def body(self):
        raise NotImplementedError()

    def __call__(self, error=None, time=None):
        """Execute benchmark
        """
        error = error or self.default_error
        time = time or self.default_time

        def result(result=None):
            if result is not None:
                results.append(float(result))
            if len(results) < self.min_count:
                return None, None
            res_mean = reduce(add, results, 0) / len(results)
            res_error = (reduce(add, ((result - res_mean) ** 2
                         for result in results), 0) / (len(results) - 1))
            return res_mean, res_error
        results = []

        @app
        def run():
            try:
                yield self.init()
                begin_time = time()
                while True:
                    start_time = time()
                    yield self.body()
                    stop_time = time()
                    res_mean, res_error = result(stop_time - start_time)
                    if res_mean is None:
                        continue
                    if (res_error / res_mean <= error or
                       (stop_time - begin_time) >= self.max_time):
                        do_return((res_mean, res_error))
            finally:
                self.dispose()

        res_mean, res_error = run()
        return (self.name, res_mean / self.factor,
                res_error / res_mean, len(results) * self.factor)

    def dispose(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False


class TextBenchmarkRunner (object):
    def __init__(self, file=None):
        self.benchs = []
        self.file = file or sys.stdout

    def add(self, bench):
        """Add benchmark
        """
        self.benchs.append(bench)

    def add_module(self, module):
        """Add benchmark from module
        """
        getattr(module, 'load_bench')(self)

    def __call__(self, error_thresh=None, timer=None):
        # run benchmarks
        lines = [('Name', 'Time', 'Count', 'Count/Time', 'Deviation',)]
        time_total = 0
        for bench in self.benchs:
            name, time, error, count = bench(error_thresh, timer)
            lines.append((name,                           # name
                         '{:.3f}s'.format(time * count),  # time
                         '{:.0f}'.format(count),          # count
                         '{:.0f}'.format(1 / time),       # count/time
                         '{:.3f}%'.format(error * 100)))  # error
            time_total += time * count
            self.file.write('.')
            self.file.flush()
        self.file.write('\n{}\n'.format('-' * 70))
        self.file.write('Ran {} benchmarks in {:.3f}s\n\n'
                        .format(len(lines), time_total))

        # calculate columns width
        widths = [0] * len(lines[0])
        for line in lines:
            for index in range(len(widths)):
                widths[index] = max(widths[index], len(line[index]))

        for index in range(len(widths)):
            widths[index] += 2

        format = '  {}\n'.format(''.join('{{:<{}}}'
                                 .format(width) for width in widths))
        lines.insert(1, tuple('-' * (width - 1) for width in widths))
        for line in lines:
            self.file.write(format.format(*line))
        self.file.write('\n')
        self.file.flush()


def main():
    from importlib import import_module
    bench_runner = TextBenchmarkRunner()
    bench_runner.add_module(import_module(__package__))
    bench_runner()

if __name__ == '__main__':
    main()



"""List monad implementation
"""
from .monad import Monad

__all__ = ('List',)


class List(tuple, Monad):
    """List monad
    """
    __slots__ = tuple()

    def __new__(cls, *vals):
        return tuple.__new__(cls, vals)

    def __init__(cls, *vals):
        pass

    @classmethod
    def from_iter(cls, iter):
        return cls(*iter)

    @classmethod
    def unit(cls, val):
        return cls(val)

    def bind(self, func):
        return List.from_iter(tuple(fval
                              for mval in self
                              for fval in func(mval).__monad__()))

    @property
    def value(self):
        return tuple(self)

    @classmethod
    def zero(cls):
        return cls()

    def plus(self, monad):
        return List.from_iter(self + monad.__monad__())

    def __reduce__(self):
        return List, tuple(self)

    def __str__(self):
        return 'List({})'.format(', '.join(repr(val) for val in self))

    __repr__ = __str__



"""Continuation monad implementation
"""
import types
from .monad import Monad
from .result import Result, callsite_banner

__all__ = ('Cont', 'Future', 'callcc',)


class Cont(Monad):
    """Continuation monad

    Haskell style continuation monad ContT {run :: (a -> r) -> r}.
    """
    __slots__ = ('run',)

    def __init__(self, run):
        self.run = run

    def __call__(self, ret=None):
        if ret is None:
            banner = callsite_banner('[continuation] error in coroutine started from')
            return self.run(lambda val: isinstance(val, Result) and val.trace(banner=banner))
        else:
            return self.run(ret)

    def __or__(self, other):
        def run(ret):
            def done_ret(val):
                if not done[0]:
                    done[0] = True
                    ret(val)
            done = [False]
            self.run(done_ret)
            other.__monad__().run(done_ret)
        return Cont(run)

    def __monad__(self):
        return self

    @classmethod
    def unit(cls, val):
        return cls(lambda ret: ret(val))

    def bind(self, func):
        return Cont(lambda ret: self.run(
                    lambda val: func(val).__monad__().run(ret)))

    def future(self):
        return Future(self)

    def __str__(self):
        return ('Cont(run:{})'.format(
                getattr(self.run, '__qualname__', self.run.__name__)
                if isinstance(self.run, types.FunctionType) else self.run))

    def __repr__(self):
        return str(self)


def callcc(func):
    """Call with current continuation

    callcc :: ((a -> Cont r b) -> Cont r a) -> Cont r a
    """
    return Cont(lambda ret: func(lambda val: Cont(
                                 lambda _: ret(val))).__monad__().run(ret))


class Future(object):
    """Future object containing future result of computation
    """
    __slots__ = ('res', 'rets',)

    def __init__(self, cont):
        self.res = None
        self.rets = []

        def ret(res):
            self.res = res
            rets, self.rets = self.rets, None
            assert rets is not None, 'continuation has been called twice'
            for ret in rets:
                ret(res)
        cont.__monad__()(ret)

    @property
    def value(self):
        if self.rets is None:
            return self.res.value if isinstance(self.res, Result) else self.res
        else:
            return Cont(self).value

    def __call__(self, ret):
        return ret(self.res) if self.rets is None else self.rets.append(ret)

    def __monad__(self):
        return Cont(self)

    def __or__(self, cont):
        return self.__monad__() | cont

    def __and__(self, cont):
        return self.__monad__() & cont

    @property
    def completed(self):
        return self.rets is None

    def __bool__(self):
        return self.rets is None

    __nonzero__ = __bool__

    def __str__(self):
        return ('Future(done:{}, value:{})'.format(self.completed, self.res))

    def __repr__(self):
        return str(self)



"""Fast and lightweight graph based state machine

Usage example:
    STATE_IDLE = 0
    STATE_EXEC = 1
    STATE_DISP = 2
    STATE_NAMES = ('idle', 'executing', 'disposed',)
    STATE_GRPAH = StateMachine.compile_graph({
        STATE_IDLE: (STATE_IDLE, STATE_EXEC, STATE_DISP),
        STATE_EXEC: (STATE_IDLE, STATE_DISP),
        STATE_DISP: (STATE_DISP,),
    })

    state = StateMachine(STATE_GRPAH, STATE_NAMES)
    state(STATE_EXEC)  # True            : change state to executing
    state(STATE_EXEC)  # raise ValueError: transition is not allowed
    state(STATE_DISP)  # True            : change state to disposed
    state(STATE_DISP)  # False           : state has not been changed
"""
import array

__all__ = ('StateMachine',)


class StateMachine(object):
    """Graph state machine
    """
    __slots__ = ('graph', 'state', 'names',)

    def __init__(self, graph, names=None):
        self.graph = graph
        self.names = names
        self.state = 0

    def __call__(self, state):
        if self.graph[self.state] & (1 << state):
            if self.state == state:
                return False
            self.state = state
            return True
        else:
            raise ValueError('invalid state transition {} -> {}'.
                             format(self.state_name(), self.state_name(state)))

    def reset(self):
        self.state = 0

    def state_name(self, state=None):
        state = self.state if state is None else state
        if self.names:
            return self.names[state]
        else:
            return str(state)

    def __str__(self):
        return '{}(state:{})'.format(type(self).__name__, self.state_name())

    def __repr__(self):
        return str(self)

    @classmethod
    def compile_graph(cls, graph_tree):
        """Create transition graph from dictionary representation

        Dictionary with state as key and available transition states as its
        value. State must be and integer. Initial state is 0.
        """
        max_state = 0
        graph = {}
        for src, dsts in graph_tree.items():
            max_state = max(max_state, src)
            for dst in dsts:
                max_state = max(max_state, dst)
                graph[src] = graph.get(src, 0) | 1 << dst

        if max_state >= 63:
            return list(graph.get(src, 0) for src in range(max_state + 1))
        else:
            return array.array('B' if max_state <= 7 else
                               'H' if max_state <= 15 else
                               'I' if max_state <= 31 else 'Q',
                               (graph.get(src, 0) for src in range(max_state + 1)))



"""Helper module (used by connection tests)

This module will cause interrupt inside Connection.do_recv if
int_function is called inside remote connection.
"""
try:
    import not_existing_module
except ImportError:
    pass

__all__ = ('remote_function',)


def int_function():
    return 'done'



"""SSL Socket Stream
"""
import socket
import errno
try:
    import ssl
except ImportError:
    ssl = None  # no SSL support

from .sock import Socket
from .buffered import BufferedStream
from ..monad import async, do_return
from ..core import POLL_READ, POLL_WRITE
from ..uniform import BrokenPipeError, BlockingErrorSet, PipeErrorSet

__all__ = ('SocketSSL', 'BufferedSocketSSL')


class SocketSSL(Socket):
    """SSL Socket Stream

    If socket has already been connected it must be wrapped with
    ssl.wrap_socket, otherwise it will be wrapped when AsyncSSLSocket.Connect
    is finished.
    """
    def __init__(self, sock, ssl_options=None, init=None, core=None):
        self.ssl_options = ssl_options or {}
        Socket.__init__(self, sock, init, core)

    @async
    def read(self, size):
        with self.reading:
            while True:
                try:
                    data = self.sock.recv(size)
                    if size and not data:
                        raise BrokenPipeError(errno.EPIPE, 'broken pipe')
                    do_return(data)
                except ssl.SSLError as error:
                    if error.args[0] != ssl.SSL_ERROR_WANT_READ:
                        raise
                except socket.error as error:
                    if error.errno not in BlockingErrorSet:
                        if error.errno in PipeErrorSet:
                            raise BrokenPipeError(error.errno, error.strerror)
                        raise
                yield self.core.poll(self.fd, POLL_READ)

    @async
    def write(self, data):
        with self.writing:
            while True:
                try:
                    do_return(self.sock.send(data))
                except ssl.SSLError as error:
                    if error.args[0] != ssl.SSL_ERROR_WANT_WRITE:
                        raise
                except socket.error as error:
                    if error.errno not in BlockingErrorSet:
                        if error.errno in PipeErrorSet:
                            raise BrokenPipeError(error.errno, error.strerror)
                        raise
                yield self.core.poll(self.fd, POLL_WRITE)

    @async
    def connect(self, address):
        yield Socket.connect(self, address)
        with self.writing, self.reading:
            self.sock = ssl.wrap_socket(self.sock, do_handshake_on_connect=False,
                                        **self.ssl_options)
            # do handshake
            while True:
                event = None
                try:
                    self.sock.do_handshake()
                    do_return(self)
                except ssl.SSLError as error:
                    if error.args[0] == ssl.SSL_ERROR_WANT_READ:
                        event = POLL_READ
                    elif error.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                        event = POLL_WRITE
                    else:
                        raise
                yield self.core.poll(self.fd, event)

    @async
    def accept(self):
        with self.reading:
            while True:
                try:
                    client, addr = self.sock.accept()
                    context = getattr(self.sock, 'context', None)
                    if context:  # python >= 3.2
                        client = context.wrap_socket(client, server_side=True)
                    else:
                        client = ssl.wrap_socket(client, server_side=True,
                                                 **self.ssl_options)
                    do_return((SocketSSL(client, self.bufsize, self.ssl_options,
                               self.core), addr))
                except socket.error as error:
                    if error.errno not in BlockingErrorSet:
                        raise
                yield self.core.poll(self.fd, POLL_READ)


class BufferedSocketSSL(BufferedStream):
    def __init__(self, sock, bufsize=None, ssl_options=None, init=None, core=None):
        BufferedStream.__init__(self, SocketSSL(sock, ssl_options, init, core), bufsize)

    def detach(self):
        return BufferedStream.detach(self).detach()

    @async
    def accept(self):
        with self.reading:
            sock, addr = yield self.base.accept()
            do_return((BufferedSocketSSL(sock.Socket, self.bufsize,
                       sock.ssl_options, True, sock.core), addr))



import textwrap
import unittest
import tempfile
import time
from . import async_test
from ..monad import Cont
from ..dispose import CompDisp
from ..process import (Process, ProcessError, PIPE, DEVNULL,
                       process_call, process_chain_call,)
from .. import PRETZEL_TEST_TIMEOUT

__all__ = ('ProcessTest',)


class ProcessTest(unittest.TestCase):
    @async_test
    def test_call(self):
        out, err, code = yield process_call(command, b'10', check=False)

        self.assertEqual(code, 117)
        self.assertEqual(out, b'02468')
        self.assertEqual(err, b'13579')

    @async_test
    def test_call_shell(self):
        shell_command = list(command)
        shell_command[-1] = '\'{}\''.format(command[-1])
        shell_command += ['| wc -c']
        out, err, code = yield process_call(shell_command, b'10', shell=True)
        self.assertEqual(code, 0)
        self.assertEqual(out, b'5\n')
        self.assertEqual(err, b'13579')

    @async_test
    def test_call_fd(self):
        with CompDisp() as dispose:
            stdin = dispose.add(tempfile.TemporaryFile())
            stdout = dispose.add(tempfile.TemporaryFile())
            stderr = dispose.add(tempfile.TemporaryFile())

            stdin.write(b'10')
            stdin.seek(0)
            out, err, code = yield process_call(command, stdin=stdin, stdout=stdout,
                                                stderr=stderr, check=False)

            self.assertEqual(code, 117)
            self.assertEqual(out, None)
            self.assertEqual(err, None)
            stdout.seek(0)
            self.assertEqual(stdout.read(), b'02468')
            stderr.seek(0)
            self.assertEqual(stderr.read(), b'13579')

    @async_test
    def test_cleanup(self):
        with (yield Process('cat', stdin=PIPE, stdout=PIPE, stderr=PIPE,
                            check=False)) as proc:
            self.assertTrue(proc.stdin.close_on_exec())
            self.assertTrue(proc.stdout.close_on_exec())
            self.assertTrue(proc.stderr.close_on_exec())
        yield proc.status

    @async_test
    def test_bad_exec(self):
        with self.assertRaises(OSError):
            print((yield process_call('does_not_exists')))

    @async_test
    def test_stress(self):
        reference = yield process_call('uname')
        procs = yield Cont.sequence(process_call('uname') for _ in range(30))
        for proc in procs:
            self.assertEqual(proc.value, reference)

    @async_test
    def test_devnull(self):
        self.assertEqual((yield process_call('cat', stdin=DEVNULL)),
                         (b'', b'', 0))

    @async_test
    def test_check_option(self):
        out, err, code = yield process_call('false', check=False)
        self.assertNotEqual(code, 0)
        with self.assertRaises(ProcessError):
            yield process_call('false', check=True)

    @async_test
    def test_kill(self):
        with (yield Process(['sleep', '10'], kill=0.5)) as proc:
            pass
        self.assertFalse(proc.status.completed)
        start = time.time()
        self.assertEqual((yield proc.status), 0)
        stop = time.time()
        self.assertTrue(stop - start < max(PRETZEL_TEST_TIMEOUT - 1, 1))

    @async_test
    def test_chain(self):
        commands = [command, ['cat'], ['wc', '-c']]

        # data in memory data
        self.assertEqual((yield process_chain_call(commands, stdin=b'10', check=False)),
                         (b'5\n', b'13579', (117, 0, 0)))

        # data in files
        with CompDisp() as dispose:
            stdin = dispose.add(tempfile.TemporaryFile())
            stdout = dispose.add(tempfile.TemporaryFile())
            stderr = dispose.add(tempfile.TemporaryFile())
            stdin.write(b'10')
            stdin.seek(0)
            result = yield process_chain_call(commands, stdin=stdin, stdout=stdout,
                                              stderr=stderr, check=False)
            self.assertEqual(result, (None, None, (117, 0, 0)))
            stdout.seek(0)
            self.assertEqual(stdout.read(), b'5\n')
            stderr.seek(0)
            self.assertEqual(stderr.read(), b'13579')

        with self.assertRaises(ProcessError):
            yield process_chain_call(commands, stdin=b'10')


command = ['python', '-c', textwrap.dedent("""\
    import sys
    for value in range(int(input())):
        if value % 2 == 1:
            sys.stderr.write(str(value))
        else:
            sys.stdout.write(str(value))
    sys.stderr.flush()
    sys.stdout.flush()
    sys.exit (117)
    """)]



import operator
import functools
import struct
from bisect import bisect_left, insort
from .serialize import Serializer

__all__ = ('StoreBlock', 'StoreAllocator',)


class StoreBlock (object):
    """Store block object

    Descriptor format:

        0       6      7 + order         64
        +-------+------+-----------------+
        | order | used | offset >> order |
        +-------+------+-----------------+

        Because data offset is always aligned by its order, first order bits
        of the offset can be used to store used size.
    """
    __slots__ = ('order', 'offset', 'used',)

    def __init__(self, order, offset, used=0):
        self.order = order
        self.offset = offset
        self.used = used

    def __lt__(self, other):
        return (self.order, self.offset) < (other.order, other.offset)

    def __eq__(self, other):
        return (self.order, self.offset) == (other.order, other.offset)

    def __hash__(self):
        return hash((self.order, self.offset))

    @property
    def size(self):
        """Size of the block
        """
        return 1 << self.order

    def to_desc(self):
        """Convert block to integer descriptor
        """
        return self.order | (self.used << 6) | (self.offset << 7)

    @classmethod
    def from_desc(cls, desc):
        """Create from integer descriptor
        """
        order = desc & 0x3f
        value = desc >> 6
        value_mask = (1 << (order + 1)) - 1
        return cls(order, (value & ~value_mask) >> 1, value & value_mask)

    def __str__(self):
        return ('StoreBlock(order:{} offset:{} used:{})'.format(
                self.order, self.offset, self.used))

    def __repr__(self):
        return str(self)


class StoreAllocator (object):
    """Store buddy allocator

    ``blocks`` is a sorted list of free blocks.
    """
    max_order = 57  # 64 - 6 (used by order) - 1 (used by size)
    desc_struct = struct.Struct('>Q')

    def __init__(self, blocks=None):
        self.blocks = blocks or [StoreBlock(self.max_order, 0)]

    def alloc(self, size):
        """Allocate block by size

        Block it is an order-offset pair.
        """
        return self.alloc_by_order((size - 1).bit_length())

    def alloc_by_order(self, order):
        """Allocate block by order

        Block it is an order-offset pair.
        """
        index = bisect_left(self.blocks, StoreBlock(order, 0))
        if index >= len(self.blocks):
            raise ValueError('out of space')
        block = self.blocks.pop(index)
        for block_order in range(order, block.order):
            insort(self.blocks,
                   StoreBlock(block_order, block.offset + (1 << block_order)))
        return StoreBlock(order, block.offset)

    def free(self, block):
        """Free previously allocated block
        """
        for order in range(block.order, self.max_order):
            buddy = StoreBlock(order, block.offset ^ (1 << order))
            index = bisect_left(self.blocks, buddy)
            if index < len(self.blocks) and buddy == self.blocks[index]:
                self.blocks.pop(index)
                block = StoreBlock(order + 1, buddy.offset
                                   if block.offset & (1 << order) else block.offset)
            else:
                self.blocks.insert(index, block)
                return
        assert not self.blocks
        self.blocks.append(block)

    @property
    def size(self):
        """Size of allocated space
        """
        return ((1 << self.max_order) - functools.reduce(operator.add,
                (block.size for block in self.blocks), 0))

    def to_stream(self, stream):
        """Save allocator to stream
        """
        descs = [block.to_desc() for block in self.blocks]
        Serializer(stream).write_struct_list(descs, self.desc_struct)
        return stream

    @classmethod
    def from_stream(cls, stream):
        """Load allocator from stream
        """
        return cls([StoreBlock.from_desc(desc) for desc in
                    Serializer(stream).read_struct_list(cls.desc_struct)])

    def __str__(self):
        return 'StoreAllocator(allocated:{})'.format(self.size)

    def __repr__(self):
        return str(self)



import random
import unittest
from ..alloc import StoreBlock, StoreAllocator

__all__ = ('StoreAllocatorTest',)


class StoreAllocatorTest(unittest.TestCase):
    """Store allocator unit tests
    """
    def test_stress(self):
        """Stress test for store allocator
        """
        alloc = StoreAllocator()
        size = 0
        blocks = []

        def reload():
            alloc.blocks = [StoreBlock.from_desc(block.to_desc()) for block in alloc.blocks]

        # fill
        for _ in range(1 << 16):
            order = random.randint(1, 10)
            size += 1 << order
            blocks.append(alloc.alloc_by_order(order))
        self.assertEqual(len(set(block.offset for block in blocks)), len(blocks))
        self.assertEqual(alloc.size, size)
        reload()

        # remove half
        for block in blocks[1 << 15:]:
            size -= block.size
            alloc.free(block)
        blocks = blocks[:1 << 15]
        self.assertEqual(len(set(block.offset for block in blocks)), len(blocks))
        self.assertEqual(alloc.size, size)
        reload()

        # add more
        for _ in range(1 << 15):
            order = random.randint(1, 10)
            size += 1 << order
            blocks.append(alloc.alloc_by_order(order))
        self.assertEqual(len(set(block.offset for block in blocks)), len(blocks))
        self.assertEqual(alloc.size, size)
        reload()

        # remove some
        for block in blocks[1 << 14:]:
            size -= block.size
            alloc.free(block)
        blocks = blocks[:1 << 14]
        self.assertEqual(len(set(block.offset for block in blocks)), len(blocks))
        self.assertEqual(alloc.size, size)
        reload()

        # remove all
        for block in blocks:
            size -= block.size
            alloc.free(block)
        self.assertEqual(size, 0)
        self.assertEqual(alloc.size, 0)
        reload()



__all__ = ('BPTreeProvider',)


class BPTreeProvider(object):
    """BPTree Provider Interface
    """
    def size(self, value=None):
        """Count of key-value pairs inside B+Tree
        """
        raise NotImplementedError()

    def depth(self, value=None):
        """Depth of B+Tree
        """
        raise NotImplementedError()

    def root(self, value=None):
        """Root node of B+Tree
        """
        raise NotImplementedError()

    def order(self):
        """Order of B+Tree Node
        """
        raise NotImplementedError()

    def node_to_desc(self, node):
        """Get nodes descriptor
        """
        raise NotImplementedError()

    def desc_to_node(self, desc):
        """Get node by its descriptor
        """
        raise NotImplementedError()

    def create(self, keys, children, is_leaf):
        """Create new node
        """
        raise NotImplementedError()

    def dirty(self, node):
        """Mark node as dirty
        """
        raise NotImplementedError()

    def release(self, node):
        """Release node
        """
        raise NotImplementedError()

    def flush(self):
        """Flush provider
        """

    def dispose(self):
        """Dispose provider
        """
        self.flush()

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __iter__(self):
        """Iterate over all nodes
        """
        stack = [self.root()]
        while stack:
            node = stack.pop()
            yield node
            if not node.is_leaf:
                for desc in node.children:
                    stack.append(self.desc_to_node(desc))

    def __str__(self):
        return ('{}(size:{}, order:{}, depth:{}, root:{})'.format(type(self).__name__,
                self.size(), self.order(), self.depth(), self.root()))

    def __repr__(self):
        return str(self)



import os
import textwrap
import unittest
import functools
import importlib
from .proxy import Remote
from ..hub import pair
from ..conn import ForkConnection, SSHConnection
from ..conn.conn import ConnectionProxy
from ..proxy import Proxy, proxify
from ...core import schedule
from ...monad import Result, async_all
from ...boot import BootLoader, boot_pack
from ...process import process_call
from ...tests import async_test
from ... import PRETZEL_POLLER, __name__ as pretzel

__all__ = ('ForkConnectionTest', 'SSHConnectionTest',)


class ForkConnectionTest(unittest.TestCase):
    conn_env = {'PRETZEL_POLLER': PRETZEL_POLLER}
    conn_type = functools.partial(ForkConnection, environ=conn_env)

    @async_test
    def test_misc(self):
        environ = self.conn_env.copy()
        environ['MYENV'] = 'MYVAL'
        with (yield self.conn_type(environ=environ)) as conn:
            # make sure process id differ
            self.assertNotEqual((yield conn(os.getpid)()), os.getpid())

            # make sure poller is the same
            self.assertEqual((yield conn(importlib.import_module)
                             (pretzel).PRETZEL_POLLER), PRETZEL_POLLER)

            # current working directory
            self.assertEqual((yield conn(os.getcwd)()), '/')

            # bad message
            with self.assertRaises(TypeError):
                yield conn.sender('bad_message')

            # result marshaling
            self.assertEqual((yield conn(Result.from_value('test_value'))),
                             'test_value')
            with self.assertRaises(RuntimeError):
                yield conn(Result.from_exception(RuntimeError()))

            # environment
            remote_environ = conn(importlib.import_module)('os').environ
            self.assertEqual((yield remote_environ.get('MYENV')), 'MYVAL')
            self.assertEqual(os.environ.get('MYENV'), None)

    @async_test
    def test_proxy(self):
        with (yield self.conn_type()) as conn:
            with (yield proxify(conn(Remote)('val'))) as proxy:
                # call
                self.assertEqual((yield proxy('test')), 'test')

                # attribute
                self.assertEqual((yield proxy.value), 'val')
                with self.assertRaises(AttributeError):
                    yield proxy.bad_attr

                # item
                self.assertEqual((yield proxy['item']), 'item_value')
                with self.assertRaises(KeyError):
                    yield proxy['bad item']

                # method
                self.assertEqual((yield proxy.method()), (yield proxy.value))
                self.assertEqual((yield proxy.method('val_new')), 'val')
                self.assertEqual((yield proxy.method()), (yield proxy.value))
                with self.assertRaises(RuntimeError):
                    yield proxy.method_error(RuntimeError())

                # asynchronous method
                async_val = (~proxy.method_async()).__monad__().future()
                self.assertFalse(async_val.completed)
                yield proxy()
                self.assertEqual((yield async_val).value, (yield proxy.value))

                with (yield proxify(proxy.value)) as value_proxy:
                    self.assertTrue(isinstance(value_proxy, Proxy))

        # process exit status
        self.assertEqual((yield conn.process.status), 0)
        self.assertFalse(conn.hub.handlers)

    @async_test
    def test_nested(self):
        """Nested connection test
        """
        pids = set()
        with (yield self.conn_type()) as c0:
            pids.add((yield c0(os.getpid)()))
            with (yield ~c0(self.conn_type)()) as c1:
                pids.add((yield c1(os.getpid)()))
                self.assertTrue(isinstance(c1, ConnectionProxy))
                with (yield ~c1(self.conn_type)()) as c2:
                    pids.add((yield c2(os.getpid)()))
        self.assertEqual(len(pids), 3)

        yield schedule()  # make sure we are not in handler
        self.assertFalse(c0.hub.handlers)

    @async_test
    def test_sender_roundtrip(self):
        r, s = pair()
        with (yield self.conn_type()) as conn:
            self.assertEqual(tuple((yield conn(s)).addr), tuple(s.addr))

    @async_test
    def test_interrupt(self):
        """Test interrupt inside Connection.do_recv
        """
        from .conn_int import int_function
        with (yield self.conn_type()) as conn:
            # We need to send two requests (second will cause interrupt)
            res0, res1 = yield async_all((conn(int_function)(), conn(int_function)()))
            self.assertEqual(res0, int_function())
            self.assertEqual(res1, int_function())

    @async_test
    def test_importer(self):
        with (yield self.conn_type()) as conn:
            self.assertEqual((yield conn(clean_path)()), [])
            self.assertEqual((yield conn(__import__)('wsgiref').__loader__ >> type),
                             BootLoader)


class SSHConnectionTest(ForkConnectionTest):
    conn_type = functools.partial(SSHConnection, host='localhost',
                                  environ=ForkConnectionTest.conn_env)

    @async_test
    def test_stdout(self):
        """Test that stdout if forwarded to stdin
        """
        test_stdout_source = textwrap.dedent("""\
            from __future__ import print_function
            from {pretzel}.app import app
            from {pretzel}.remoting import SSHConnection
            @app
            def main():
                with (yield SSHConnection('localhost')) as conn:
                    yield conn(print)('done', end='')
            main()
            """.format(pretzel=pretzel))
        result = yield process_call(['python', '-c', boot_pack(test_stdout_source)])
        self.assertEqual(result, (b'', b'done', 0))


def clean_path():  # pragma: no cover
    """Clean system path to force use of connection importer

    Called from remote connection.
    """
    import sys
    del sys.path[:]
    return sys.path



"""Synchronous version of BufferedStream's serializer
"""
import struct

__slots__ = ('Serializer',)


class Serializer(object):
    """Stream serializer

    It is synchronous version of BufferedStream's serializer.
    """
    size_struct = struct.Struct('>I')

    def __init__(self, stream):
        self.stream = stream

    def read_bytes(self):
        """Read bytes object
        """
        return (self.stream.read(self.size_struct.unpack
               (self.stream.read(self.size_struct.size))[0]))

    def write_bytes(self, bytes):
        """Write bytes object to buffer
        """
        self.stream.write(self.size_struct.pack(len(bytes)))
        self.stream.write(bytes)

    def read_struct_list(self, struct, complex=None):
        """Read list of structures
        """
        struct_data = self.read_bytes()
        if complex:
            return [struct.unpack(struct_data[offset:offset + struct.size])
                    for offset in range(0, len(struct_data), struct.size)]
        else:
            return [struct.unpack(struct_data[offset:offset + struct.size])[0]
                    for offset in range(0, len(struct_data), struct.size)]

    def write_struct_list(self, struct_list, struct, complex=None):
        """Write list of structures to buffer
        """
        self.stream.write(self.size_struct.pack(len(struct_list) * struct.size))
        if complex:
            for struct_target in struct_list:
                self.stream.write(struct.pack(*struct_target))
        else:
            for struct_target in struct_list:
                self.stream.write(struct.pack(struct_target))

    def read_bytes_list(self):
        """Read array of bytes
        """
        return [self.stream.read(size)
                for size in self.read_struct_list(self.size_struct)]

    def write_bytes_list(self, bytes_list):
        """Write bytes array object to buffer
        """
        self.write_struct_list([len(bytes) for bytes in bytes_list], self.size_struct)
        for bytes in bytes_list:
            self.stream.write(bytes)



"""Store data inside some flat store (file, memory)

Data can be named and unnamed. Named data addressed by its name.
Unnamed data addressed by its descriptor which can change if data changed.
Descriptor is an unsigned 64-bit integer.
"""
import io
import os
import struct
from .alloc import StoreBlock, StoreAllocator
from .serialize import Serializer
from ..dispose import CompDisp

__all__ = ('Store', 'StreamStore', 'FileStore',)


class Store(object):
    """Store data inside some flat store (file, memory)

    Data can be named and unnamed. Named data addressed by its name.
    Unnamed data addressed by its descriptor which can change if data changed.
    Descriptor is an unsigned 64-bit integer.
    """

    header_struct = struct.Struct('>QQ')
    desc_struct = struct.Struct('>Q')

    def __init__(self, offset=None):
        self.disps = CompDisp()
        offset = offset or 0
        self.offset = offset + self.header_struct.size

        header = self.load_by_offset(offset, self.header_struct.size)
        self.alloc_desc, self.names_desc = ((0, 0) if not header else
                                            self.header_struct.unpack(header))

        # allocator
        if self.alloc_desc:
            self.alloc = StoreAllocator.from_stream(io.BytesIO(self.load(self.alloc_desc)))
        else:
            self.alloc = StoreAllocator()

        # names
        if self.names_desc:
            serialzer = Serializer(io.BytesIO(self.load(self.names_desc)))
            self.names = dict(zip(
                serialzer.read_bytes_list(),                    # names
                serialzer.read_struct_list(self.desc_struct)))  # descriptors
        else:
            self.names = {}

    def load(self, desc):
        """Load data by descriptor
        """
        if not desc:
            return b''
        block = StoreBlock.from_desc(desc)
        return self.load_by_offset(self.offset + block.offset, block.used)

    def load_by_name(self, name):
        """Load data by name
        """
        return self.load(self.names.get(name, 0))

    def load_by_offset(self, offset, size):
        """Load data by offset and size
        """
        raise NotImplementedError()

    def __getitem__(self, name):
        """Load data by name
        """
        return self.load_by_name(name)

    def save(self, data, desc=None):
        """Save data by descriptor

        Try to save data inside space pointed by descriptor and
        if its not enough allocate new space. Returns descriptor of saved data
        """
        if not data:
            return 0

        block = self.reserve_block(len(data), desc)
        self.save_by_offset(self.offset + block.offset, data)
        return block.to_desc()

    def save_by_name(self, name, data):
        """Save data by name
        """
        if not data:
            self.delete_by_name(name)
        else:
            self.names[name] = self.save(data, self.names.get(name))
        return data

    def save_by_offset(self, offset, data):
        """Save data by offset
        """
        raise NotImplementedError()

    def __setitem__(self, name, data):
        """Save data by name
        """
        self.save_by_name(name, data)

    def delete(self, desc):
        """Delete data by descriptor

        Free space occupied by data pointed by descriptor
        """
        if not desc:
            return
        self.alloc.free(StoreBlock.from_desc(desc))

    def delete_by_name(self, name):
        """Delete data by name
        """
        desc = self.names.pop(name, None)
        if desc:
            self.delete(desc)

    def __delitem__(self, name):
        """Delete data by name
        """
        self.delete_by_name(name)

    def reserve(self, size, desc=None):
        """Reserve space without actually writing anything in it

        Returns store's block descriptor.
        """
        return self.reserve_block(size, desc).to_desc()

    def reserve_block(self, size, desc=None):
        """Reserve space without actually writing anything in it

        Return store block.
        """
        if desc:
            block = StoreBlock.from_desc(desc)
            if block.size >= size:
                block.used = size
                return block
            self.alloc.free(block)
        block = self.alloc.alloc(size)
        block.used = size
        return block

    def flush(self):
        """Flush current state
        """
        # names
        if self.names:
            sr = Serializer(io.BytesIO())
            sr.write_bytes_list(tuple(self.names.keys()))
            sr.write_struct_list(tuple(self.names.values()), self.desc_struct)
            self.names_desc = self.save(sr.stream.getvalue(), self.names_desc)
        else:
            self.delete(self.names_desc)
            self.names_desc = 0

        # Check if nothing is allocated of the only thing allocated is
        # allocator itself.
        if self.alloc.size - (StoreBlock.from_desc(self.alloc_desc).size if self.alloc_desc else 0):
            while True:
                alloc_state = self.alloc.to_stream(io.BytesIO()).getvalue()
                self.alloc_desc, alloc_desc = self.save(alloc_state, self.alloc_desc), self.alloc_desc
                if self.alloc_desc == alloc_desc:
                    break
        else:
            alloc_desc, self.alloc_desc = self.alloc_desc, 0
            self.delete(alloc_desc)
            assert not self.alloc.size, 'Allocator is broken'

        # header
        self.save_by_offset(
            self.offset - self.header_struct.size,
            self.header_struct.pack(self.alloc_desc, self.names_desc))

    @property
    def size(self):
        """Total space used excluding internal storage data
        """
        size = 0
        if self.alloc_desc:
            size += StoreBlock.from_desc(self.alloc_desc).size
        if self.names_desc:
            size += StoreBlock.from_desc(self.names_desc).size
        for desc in self.names.values():
            size += StoreBlock.from_desc(desc).size
        return self.alloc.size - size

    def create_cell(self, name):
        """Cell object

        Cell is a callable object. When called with argument sets argument as new
        stored value and returns stored value. Otherwise just returns stored value.
        """
        def cell(value=None):
            """Get/Set value

            If value is not set returns stored value, otherwise sets value as
            stored value.
            """
            return(self.load_by_name(name) if value is None else
                   self.save_by_name(name, value))
        name = name if isinstance(name, bytes) else name.encode()
        return cell

    def create_map(self, name, order=None, key_type=None, value_type=None, compress=None):
        """Create name mapping (B+Tree)
        """
        from .map import StoreMap
        cell = self.create_cell('__map:{}'.format(name))
        mapping = StoreMap(self, cell, order, key_type, value_type, compress)
        self.disps.add(mapping)
        return mapping

    def create_stream(self, name, bufsize=None, compress=None):
        """Create stream object with store back-end.
        """
        from .stream import StoreStream
        cell = self.create_cell('__stream:{}'.format(name))
        stream = StoreStream(self, cell, bufsize, compress)
        self.disps.add(stream)
        return stream

    def dispose(self):
        """Flush and Close
        """
        self.disps()
        self.flush()

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return '{}(size:{}, names:{})'.format(type(self).__name__,
                                              self.size, len(self.names))

    def __repr__(self):
        return str(self)


class StreamStore(Store):
    """Stream based store
    """
    def __init__(self, stream, offset=None):
        self.stream = stream
        Store.__init__(self, offset)

    def save_by_offset(self, offset, data):
        self.stream.seek(offset)
        return self.stream.write(data)

    def load_by_offset(self, offset, size):
        self.stream.seek(offset)
        return self.stream.read(size)

    def flush(self):
        Store.flush(self)
        self.stream.flush()


class FileStore(StreamStore):
    """File based store
    """
    def __init__(self, path, mode=None, offset=None):
        mode = mode or 'r'
        self.mode = mode
        if mode == 'r':
            filemode = 'rb'
        elif mode == 'w':
            filemode = 'r+b'
        elif mode == 'c':
            if not os.path.lexists(path):
                filemode = 'w+b'
            else:
                filemode = 'r+b'
        elif mode == 'n':
            filemode = 'w+b'
        else:
            raise ValueError('Unknown mode: {}'.format(mode))
        StreamStore.__init__(self, io.open(path, filemode, buffering=0), offset)

    def flush(self):
        if self.mode != 'r':
            StreamStore.flush(self)

    def dispose(self):
        StreamStore.dispose(self)
        self.stream.close()

    def __str__(self):
        return ('FileStream(size:{}, names:{}, mode:{}, path:{})'.format
               (self.size, len(self.names), self.mode, self.stream.name))



"""Socket stream

Wraps python socket object and provides stream interface.
"""
import socket
import errno

from .file import File
from .buffered import BufferedStream
from ..monad import async, do_return
from ..core import POLL_READ, POLL_WRITE
from ..uniform import BrokenPipeError, BlockingErrorSet, PipeErrorSet

__all__ = ('Socket', 'BufferedSocket',)


class Socket(File):
    """Socket stream
    """
    def __init__(self, sock, init=None, core=None):
        self.sock = sock  # used by blocking method
        File.__init__(self, sock.fileno(), closefd=False, init=bool(init), core=core)

    @async
    def read(self, size):
        with self.reading:
            while True:
                try:
                    data = self.sock.recv(size)
                    if size and not data:
                        raise BrokenPipeError(errno.EPIPE, 'broken pipe')
                    do_return(data)
                except socket.error as error:
                    if error.errno not in BlockingErrorSet:
                        if error.errno in PipeErrorSet:
                            raise BrokenPipeError(error.errno, error.strerror)
                        raise
                yield self.core.poll(self.fd, POLL_READ)

    @async
    def write(self, data):
        with self.writing:
            while True:
                try:
                    do_return(self.sock.send(data))
                except socket.error as error:
                    if error.errno not in BlockingErrorSet:
                        if error.errno in PipeErrorSet:
                            raise BrokenPipeError(error.errno, error.strerror)
                        raise
                yield self.core.poll(self.fd, POLL_WRITE)

    @async
    def connect(self, address):
        with self.initing:
            try:
                self.sock.connect(address)
                do_return(self)
            except socket.error as error:
                if error.errno not in BlockingErrorSet:
                    raise
            yield self.core.poll(self.fd, POLL_WRITE)
            do_return(self)

    @async
    def accept(self):
        with self.reading:
            while True:
                try:
                    client, addr = self.sock.accept()
                    do_return((Socket(client, init=True, core=self.core), addr))
                except socket.error as error:
                    if error.errno not in BlockingErrorSet:
                        raise
                yield self.core.poll(self.fd, POLL_READ)

    def bind(self, address):
        with self.initing:
            return self.sock.bind(address)

    def listen(self, backlog):
        if self.disposed:
            raise ValueError('socket is disposed')
        return self.sock.listen(backlog)

    def shutdown(self, how):
        if self.disposed:
            raise ValueError('socket is disposed')
        return self.sock.shutdown(how)

    def dispose(self):
        if File.dispose(self):
            sock, self.sock = self.sock, None
            if sock:
                sock.close()
            return True
        return False

    @async
    def detach(self):
        sock, self.sock = self.sock, None
        if not self.dispose():
            raise ValueError('socket is disposed')
        self.blocking(True)
        return sock

    def blocking(self, enable=None):
        if enable is None:
            return self.sock.gettimeout() != 0.0
        self.sock.setblocking(enable)
        return enable


class BufferedSocket (BufferedStream):
    def __init__(self, sock, bufsize=None, init=None, core=None):
        BufferedStream.__init__(self, Socket(sock, init=init, core=core), bufsize)

    def detach(self):
        return BufferedStream.detach(self).detach()

    @async
    def accept(self):
        with self.reading:
            sock, addr = yield self.base.accept()
            do_return((BufferedSocket(sock.sock, self.bufsize, True, sock.core), addr))



"""Core is an asynchronous execution loop
"""
import os
import sys
import errno
import signal
import itertools
import threading
from time import time
from heapq import heappush, heappop
if sys.version_info[0] > 2:
    from _thread import get_ident
else:  # pragma: no cover
    from thread import get_ident

from .poll import Poller, POLL_ERROR, POLL_READ, POLL_WRITE, POLL_DISCONNECT
from ..uniform import BrokenPipeError, ConnectionError, CanceledError, BlockingErrorSet
from ..monad import Result, async, async_block, do_done
from ..state_machine import StateMachine
from ..dispose import CompDisp

__all__ = ('Core',)

"""
Core's maximum timeout for sleeping inside poll. It easier to manipulate positive
timeout then doing branching with negative values. And waking up once in a while
won't hurt.
"""
CORE_TIMEOUT = 3600.0


class Core(object):
    """Core object

    Asynchronous I/O and Timer execution loop. Executes until all requested
    asynchronous operation are completed or when object itself is disposed.
    All interaction with the Core must be done from that Core's thread,
    exception are schedule() and wake().
    """

    STATE_IDLE = 0
    STATE_EXEC = 1
    STATE_DISP = 2
    STATE_NAMES = ('idle', 'executing', 'disposed',)
    STATE_GRPAH = StateMachine.compile_graph({
        STATE_IDLE: (STATE_IDLE, STATE_EXEC, STATE_DISP),
        STATE_EXEC: (STATE_IDLE, STATE_DISP),
        STATE_DISP: (STATE_DISP,),
    })

    inst_lock = threading.RLock()
    inst_local = threading.local()
    inst_main = None

    def __init__(self, poller=None):
        self.tick = 0
        self.state = StateMachine(self.STATE_GRPAH, self.STATE_NAMES)

        self.files_queue = {}
        self.time_queue = TimeQueue()
        self.sched_queue = SchedQueue(self)
        self.proc_queue = ProcQueue(self)

        self.thread_ident = None
        self.poller = Poller.from_name(poller)
        self.waker = Waker(self)

        def dispose_core():
            if not self.state(self.STATE_DISP):
                return
            try:
                exc = CanceledError('core has been disposed')
                files_queue, self.files_queue = self.files_queue, {}
                for file in files_queue.values():
                    file.dispose(exc)
                self.time_queue.dispose(exc)
                self.sched_queue.dispose(exc)
                self.proc_queue.dispose(exc)
            finally:
                self.waker.dispose()
                self.poller.dispose()
        self.dispose = CompDisp()
        self.dispose.add_action(dispose_core)

    @classmethod
    def main(cls, inst=None):
        """Main core instance
        """
        with cls.inst_lock:
            if inst is None:
                if cls.inst_main is None:
                    cls.inst_main = Core()
                inst = cls.inst_main
            else:
                cls.inst_main = inst
        return inst

    @classmethod
    def local(cls, inst=None):
        """Thread local instance
        """
        if inst is None:
            inst = getattr(cls.inst_local, 'inst', None)
            if inst is None:
                return cls.local(cls.main())
        else:
            cls.inst_local.inst = inst
        return inst

    @async
    def sleep(self, delay):
        """Sleep

        Interrupt current coroutine for specified amount of time
        """
        if self.state.state == self.STATE_DISP:
            raise CanceledError('core is disposed')
        do_done(self.time_queue.on(time() + delay))

    @async
    def sleep_until(self, when):
        """Sleep until

        Interrupt current coroutine until specified time is reached
        """
        if self.state.state == self.STATE_DISP:
            raise CanceledError('core is disposed')
        do_done(self.time_queue.on(when))

    @async
    def poll(self, fd, mask):
        """Poll file descriptor

        Poll file descriptor for events specified by mask. If mask is None then
        specified descriptor is unregistered and all pending events are resolved
        with BrokenPipeError, otherwise future is resolved with bitmap of the
        events happened on file descriptor or error if any.
        """
        if mask is not None and self.state.state == self.STATE_DISP:
            raise CanceledError('core is disposed')
        file = self.files_queue.get(fd)
        if file is None:
            file = FileQueue(fd, self.poller)
            self.files_queue[fd] = file
        do_done(file.on(mask))

    @async
    def schedule(self):
        """Schedule continuation to be executed on this core

        Scheduled continuation will be executed on next iteration circle. This
        function can be called from different thread, but not from signal handler
        as heappush used by time_queue is not reentrant. Returns this core object.
        """
        if self.state.state == self.STATE_DISP:
            raise CanceledError('core is disposed')
        if self.thread_ident == get_ident():
            do_done(self.time_queue.on(0).then_val(self))
        else:
            do_done(self.sched_queue.on())

    @async
    def waitpid(self, pid):
        """Wait pid

        Schedule continuation to be executed when process with pid is terminated.
        """
        if self.state.state == self.STATE_DISP:
            raise CanceledError('core is disposed')
        do_done(self.proc_queue.on(pid))

    def wake(self):
        if self.thread_ident != get_ident():
            self.waker()

    def __call__(self, dispose=False):
        return self.start()

    def start(self, dispose=False):
        """Start core execution
        """
        self.state(self.STATE_EXEC)
        try:
            for _ in self.iterator():
                if self.state.state != self.STATE_EXEC:
                    break
        finally:
            if self.state.state != self.STATE_DISP:
                if dispose:
                    self.dispose()
                else:
                    self.state(self.STATE_IDLE)

    def stop(self):
        """Stop core execution
        """
        return self.state(self.STATE_IDLE)

    def iterator(self, block=True):
        """Core's iterator

        Starts new iteration loop. Returns generator object which yield at the
        beginning of each iteration.
        """
        if self.state.state == self.STATE_DISP:
            raise RuntimeError('core is disposed')

        top_level = False
        wakeup_fd = None
        try:
            # Thread identity is used by wake() to make sure call to waker is
            # really needed. And it also used to make sure core is iterating only
            # on one thread.
            with self.inst_lock:
                if self.thread_ident is None:
                    self.thread_ident = get_ident()
                    top_level = True
                elif self.thread_ident != get_ident():
                    raise ValueError('core has already being run on a different thread')

            # Install signal wakeup file descriptor
            try:
                wakeup_fd = signal.set_wakeup_fd(self.waker.fileno())
            except ValueError:
                pass # non-main thread

            timer = self.time_queue
            files = self.files_queue
            sched = self.sched_queue
            proc = self.proc_queue
            poll = self.poller.poll

            events = tuple()
            while True:
                for fd, event in events:
                    files[fd](event)
                timer(time())
                sched()
                proc()

                # Yield control to check conditions before blocking (Core has been
                # stopped or desired future resolved). If there is no file
                # descriptors registered and timeout is negative poller will raise
                # StopIteration and break this loop.
                yield

                events = (poll(0) if not block else
                          poll(min(timer.timeout(time()), sched.timeout())))
                self.tick += 1
        finally:
            if wakeup_fd is not None:
                signal.set_wakeup_fd(wakeup_fd)
            if top_level:
                self.thread_ident = None

    def __iter__(self):
        return iter(self.iterator())

    @property
    def disposed(self):
        return self.state.state == self.STATE_DISP

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return '{}(state:{}, tick:{})'.format(type(self).__name__,
                                              self.state.state_name(), self.tick)

    def __repr__(self):
        return str(self)


class TimeQueue(object):
    """Time queue

    Schedule continuation to be executed when specified time is reached.
    """
    __slots__ = ('queue', 'uid',)

    def __init__(self):
        self.uid = itertools.count()  # used to distinguish simultaneous continuations
        self.queue = []

    def on(self, when):
        return async_block(lambda ret:
                           heappush(self.queue, (when, next(self.uid), ret)))

    def __call__(self, time_now):
        if not self.queue:
            return
        resolved = []
        while self.queue:
            time_when, _, ret = self.queue[0]
            if time_when > time_now:
                break
            resolved.append(heappop(self.queue))
        for time_when, _, ret in resolved:
            ret(time_when)

    def timeout(self, now):
        if self.queue:
            return min(CORE_TIMEOUT, max(0, self.queue[0][0] - now))
        else:
            return CORE_TIMEOUT

    def dispose(self, exc=None):
        error = Result.from_exception(exc or CanceledError('time queue has been disposed'))
        queue, self.queue = self.queue, []
        for when, _, ret in queue:
            ret(error)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __len__(self):
        return len(self.queue)

    def __str__(self):
        return 'TimeQueue(len:{})'.format(len(self))

    def __repr__(self):
        return str(self)


class FileQueue(object):
    """File queue

    Schedule continuation to be executed when specified file descriptor
    conditions are met.
    """
    __slots__ = ('fd', 'poller', 'mask', 'handlers',)

    def __init__(self, fd, poller):
        self.fd = fd
        self.poller = poller
        self.mask = 0
        self.handlers = []

    def on(self, mask):
        @async_block
        def cont(ret):
            if mask is None:
                self.dispose(BrokenPipeError(errno.EPIPE, 'detached from core'))
                ret(None)
            elif not mask:
                raise ValueError('empty mask')
            elif self.mask & mask:
                raise ValueError('intersecting mask')
            else:
                if self.mask:
                    self.poller.modify(self.fd, self.mask | mask)
                else:
                    self.poller.register(self.fd, mask)
                self.mask |= mask
                self.handlers.append((mask, ret))
        return cont

    def off(self, mask):
        if not (mask & self.mask):
            return []

        disable, enable = [], []
        for msk, ret in self.handlers:
            if msk & mask:
                self.mask &= ~msk
                disable.append(ret)
            else:
                enable.append((msk, ret))
        if self.mask:
            self.poller.modify(self.fd, self.mask)
        else:
            self.poller.unregister(self.fd)

        self.handlers = enable
        return disable

    def __call__(self, mask):
        if mask & ~POLL_ERROR:
            for ret in self.off(mask):
                ret(mask)
        else:
            exc = (BrokenPipeError(errno.EPIPE, 'broken pipe') if mask & POLL_DISCONNECT else
                   ConnectionError())
            self.dispose(exc)

    def dispose(self, exc=None):
        error = Result.from_exception(exc or CanceledError('file queue has been disposed'))
        for ret in self.off(self.mask):
            ret(error)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        flags = ','.join(name for flag, name in ((POLL_READ, 'read'),
                        (POLL_WRITE, 'write'), (POLL_ERROR, 'error')) if self.mask & flag)
        return '{}(fd:{}, flags:{})'.format(type(self).__name__, self.fd, flags)

    def __repr__(self):
        return str(self)


class SchedQueue(object):
    """Scheduler queue

    Schedules continuation to be executed on specified core
    """
    def __init__(self, core):
        self.core = core
        self.rets = []
        self.rets_lock = threading.RLock()

    def on(self):
        @async_block
        def cont(ret):
            with self.rets_lock:
                self.rets.append(ret)
            self.core.wake()
        return cont

    def __call__(self):
        with self.rets_lock:
            rets, self.rets = self.rets, []
        for ret in rets:
            ret(self.core)

    def timeout(self):
        return 0 if self.rets else CORE_TIMEOUT

    def dispose(self, exc=None):
        error = Result.from_exception(exc or CanceledError('scheduler queue has been disposed'))
        with self.rets_lock:
            rets, self.rets = self.rets, []
        for ret in self.rets:
            ret(error)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __len__(self):
        return len(self.rets)

    def __str__(self):
        return 'SchedQueue(len:{})'.format(len(self))

    def __repr__(self):
        return str(self)


class ProcQueue(object):
    """Process queue

    Schedule continuation to be executed when process with specified pid
    has been terminated.
    """
    current = [None]
    current_lock = threading.RLock()

    def __init__(self, core):
        self.core = core
        self.pids = {}
        self.pending = False

    def init(self):
        if self.current[0] == self:
            return
        elif self.current[0] is None:
            with self.current_lock:
                if self.current[0] is None:
                    self.current[0] = self
                    signal.signal(signal.SIGCHLD,
                                  lambda *_: setattr(self, 'pending', True))
                    return
        raise ValueError('process queue can only be used by single core')

    def on(self, pid):
        @async_block
        def cont(ret):
            pid_done, status = os.waitpid(pid, os.WNOHANG)
            if pid_done == pid:
                ret(os.WEXITSTATUS(status))
            else:
                if self.pids.get(pid):
                    raise ValueError('pid {} has already being waited'.format(pid))
                self.pids[pid] = ret
        self.init()
        return cont

    def __call__(self):
        pending, self.pending = self.pending, False
        if not pending:
            return

        resolved = []
        for pid, ret in tuple(self.pids.items()):
            try:
                pid_done, status = os.waitpid(pid, os.WNOHANG)
                if pid_done == pid:
                    self.pids.pop(pid, None)
                    resolved.append((ret, os.WEXITSTATUS(status)))
            except OSError:
                self.pids.pop(pid, None)
                resolved.append((ret, Result.from_current_error()))
        for ret, status in resolved:
            ret(status)

    def dispose(self, exc=None):
        if self.current[0] == self:
            signal.signal(signal.SIGCHLD, signal.SIG_DFL)
            self.current[0] = None
        error = Result.from_exception(exc or CanceledError('process queue has been disposed'))
        pids, self.pids = self.pids, {}
        for ret in self.pids.values():
            ret(error)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __len__(self):
        return len(self.pids)

    def __str__(self):
        return 'ProcQueue(len:{})'.format(len(self))

    def __repr__(self):
        return str(self)


class Waker(object):
    """Core waker
    """
    def __init__(self, core):
        self.core = core
        self.reader, self.writer = os.pipe()

        from ..stream.file import fd_blocking, fd_close_on_exec
        fd_blocking(self.reader, False)
        fd_blocking(self.writer, False)
        fd_close_on_exec(self.reader, True)
        fd_close_on_exec(self.writer, True)

        @async
        def consumer():
            """Consumer coroutine"""
            try:
                while True:
                    try:
                        data = os.read(self.reader, 65536)
                        if not data:
                            break
                    except OSError as error:
                        if error.errno not in BlockingErrorSet:
                            break
                    yield self.core.poll(self.reader, POLL_READ)
            except (CanceledError, BrokenPipeError):
                pass
            finally:
                self.dispose()
        consumer()()

    def fileno(self):
        return self.writer

    def __call__(self):
        try:
            os.write(self.writer, b'\x00')
        except OSError as error:
            if error.errno not in BlockingErrorSet:
                raise

    def dispose(self):
        reader, self.reader = self.reader, -1
        if reader >= 0:
            os.close(reader)
            self.core.poll(reader, None)
        writer, self.writer = self.writer, -1
        if writer >= 0:
            os.close(writer)
            self.core.poll(writer, None)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False



"""Fork connection

Connection with forked and exec-ed process via two pipes.
"""
import os
import sys
from .stream import StreamConnection
from ..importer import Importer
from ...process import Process, ProcessPipe, PIPE
from ...monad import async
from ...stream import BufferedFile
from ...core import Core
from ...boot import BootImporter, __name__ as boot_name

__all__ = ('ForkConnection',)


class ForkConnection(StreamConnection):
    """Fork connection

    Connection with forked and exec-ed process via two pipes.
    """
    def __init__(self, command=None, environ=None, bufsize=None, hub=None, core=None):
        StreamConnection.__init__(self, hub=hub, core=core)
        self.bufsize = bufsize
        self.command = [sys.executable, '-'] if command is None else command
        self.environ = environ
        self.process = None

    @async
    def do_connect(self, target):
        """Fork connect implementation

        Target is pickle-able and call-able which will be called upon successful
        connection with connection this as its only argument.
        """
        # pipes
        reader = self.dispose.add(ProcessPipe(True, bufsize=self.bufsize, core=self.core))
        writer = self.dispose.add(ProcessPipe(False, bufsize=self.bufsize, core=self.core))

        # process
        def preexec():  # pragma: no cover
            reader()
            writer()
            os.chdir('/')
            os.setsid()
            if self.environ:
                os.environ.update(self.environ)

        self.process = self.dispose.add(Process(self.command,
                                        stdin=PIPE, preexec=preexec,
                                        bufsize=self.bufsize, core=self.core))
        yield self.process  # exec-ed

        # send payload
        payload = (BootImporter.from_modules().bootstrap
                  (fork_conn_init, writer.child_fd, reader.child_fd,
                   self.bufsize).encode())
        self.process.stdin.write_schedule(payload)
        yield self.process.stdin.flush_and_dispose()

        reader_stream = self.dispose.add(reader())
        writer_stream = self.dispose.add(writer())
        yield StreamConnection.do_connect(self, (reader_stream, writer_stream))

        # install importer
        self.dispose.add((yield Importer.create_remote(self)))
        self.module_map['_boot'] = boot_name

        # update name
        self.flags['pid'] = self.process.pid
        self.flags['type'] = 'fork'


def fork_conn_init(reader_fd, writer_fd, bufsize):  # pragma: no cover
    """Fork connection initialization function
    """
    with Core.local() as core:
        # initialize connection
        conn = StreamConnection(core=core)
        conn.flags['pid'] = os.getpid()
        conn.flags['type'] = 'fork'
        conn.dispose.add_action(lambda: core.schedule()(lambda _: core.dispose()))

        # connect
        reader = BufferedFile(reader_fd, bufsize=bufsize, core=core)
        reader.close_on_exec(True)
        writer = BufferedFile(writer_fd, bufsize=bufsize, core=core)
        writer.close_on_exec(True)
        conn.connect((reader, writer))()

        # execute core
        if not core.disposed:
            core()



"""Proxy monad
"""

__all__ = ('Proxy',)


class Proxy(object):
    """Proxy monad

    All operation on this proxy object are executed inside context of the monad
    and on bound monad value.
    """
    __slots__ = ('__monad',)

    def __init__(self, monad):
        self.__monad = monad.__monad__()

    def __monad__(self):
        return self.__monad

    def __map(self, func):
        monad = self.__monad
        return type(self)(monad.bind(lambda val: monad.unit(func(val))))

    def __call__(self, *a, **kw):
        return self.__map(lambda val: val(*a, **kw))

    ## attributes
    def __getattr__(self, name):
        return self.__map(lambda val: getattr(val, name))

    def __delattr__(self, name):
        return self.__map(lambda val: val.__delattr__(name))

    ## items
    def __getitem__(self, item):
        return self.__map(lambda val: val[item])

    def __delitem__(self, item):
        return self.__map(lambda val: val.__delitem__(item))

    def __contains__(self, item):
        return self.__map(lambda val: val.__contains__(item))

    ## compare
    def __lt__(self, other):
        return self.__map(lambda val: val.__lt__(other))

    def __le__(self, other):
        return self.__map(lambda val: val.__le__(other))

    def __eq__(self, other):
        return self.__map(lambda val: val.__eq__(other))

    def __ne__(self, other):
        return self.__map(lambda val: val.__ne__(other))

    def __gt__(self, other):
        return self.__map(lambda val: val.__gt__(other))

    def __ge__(self, other):
        return self.__map(lambda val: val.__ge__(other))

    ## arithmetic
    def __add__(self, other):
        return self.__map(lambda val: val.__add__(other))

    def __sub__(self, other):
        return self.__map(lambda val: val.__sub__(other))

    def __mul__(self, other):
        return self.__map(lambda val: val.__mul__(other))

    def __truediv__(self, other):
        return self.__map(lambda val: val.__truediv__(other))

    def __floordiv__(self, other):
        return self.__map(lambda val: val.__floordiv__(other))

    def __mod__(self, other):
        return self.__map(lambda val: val.__mod__(other))

    def __divmod__(self, other):
        return self.__map(lambda val: val.__divmod__(other))

    def __pow__(self, other):
        return self.__map(lambda val: val.__pow__(other))

    def __lshift__(self, other):
        return self.__map(lambda val: val.__lshift__(other))

    def __rshift__(self, other):
        return self.__map(lambda val: val.__rshift__(other))

    def __and__(self, other):
        return self.__map(lambda val: val.__and__(other))

    def __xor__(self, other):
        return self.__map(lambda val: val.__xor__(other))

    def __or__(self, other):
        return self.__map(lambda val: val.__or__(other))

    def __neg__(self):
        return self.__map(lambda val: val.__neg__())

    def __pos__(self):
        return self.__map(lambda val: val.__pos__())

    def __abs__(self):
        return self.__map(lambda val: val.__abs__())

    def __invert__(self):
        return self.__map(lambda val: val.__invert__())

    ## scope
    def __enter__(self):
        return self.__map(lambda val: val.__enter__())

    def __exit__(self, et, eo, tb):
        self.__map(lambda val: val.__exit__(et, eo, tb))
        return False

    ## representation
    def __str__(self):
        return str(self.__monad)

    def __repr__(self):
        return repr(self.__monad)



import io
import os
import itertools
import unittest
from ..file import File
from ...monad import async
from ...uniform import BrokenPipeError
from ...core import schedule
from ...tests import async_test

__all__ = ('FileTest',)


class FileTest(unittest.TestCase):
    BLOCK = b'X' * 1024

    @async_test
    def test(self):
        received = io.BytesIO()
        reader_fd, writer_fd = os.pipe()

        @async
        def reader_coro():
            with File(reader_fd) as reader:
                try:
                    while True:
                        received.write((yield reader.read(1024)))
                except BrokenPipeError:
                    pass
        reader_future = reader_coro().future()
        reader_future.__monad__()()  # error trace if any
        if reader_future.completed:
            yield reader_future

        with File(writer_fd) as writer:
            # fill file buffer
            size = 0
            for _ in itertools.count():
                write = writer.write(self.BLOCK).future()
                if not write.completed:
                    size += yield write
                    break
                else:
                    size += write.value

            # drain file
            while received.tell() < size:
                yield schedule()
            self.assertEqual(received.getvalue(), b'X' * size)
            received.seek(0)
            received.truncate(0)

            yield writer.write(b'one')
            yield schedule()

            yield writer.write(b', two')
            yield schedule()

            yield writer.write(b', three')
            yield schedule()

        yield reader_future
        self.assertEqual(received.getvalue(), b'one, two, three')



"""B+Tree implementation
"""
import sys
from bisect import bisect, bisect_left
from operator import itemgetter
from collections import MutableMapping
if sys.version_info[0] < 3:
    from itertools import imap as map

__all__ = ('BPTree', 'BPTreeNode', 'BPTreeLeaf',)


class Nothing(object):
    def __str__(self):
        return 'Nothing'

    def __repr__(self):
        return str(self)

nothing = Nothing()


class BPTree(MutableMapping):
    """B+Tree
    """
    def __init__(self, provider):
        self.provider = provider

    def __get(self, key, value=nothing):
        """Get value associated with key

        Returns value associated with key if any, otherwise value argument if
        it is set, else raises KeyValue exception.
        """

        desc_to_node = self.provider.desc_to_node

        # find leaf
        node = self.provider.root()
        for _ in range(self.provider.depth() - 1):
            node = desc_to_node(node.children[bisect(node.keys, key)])

        # find key
        index = bisect_left(node.keys, key)
        if index >= len(node.keys) or key != node.keys[index]:
            if value is self.nothing:
                raise KeyError(key)
            return value

        return node.children[index]

    def __get_range(self, low_key=None, high_key=None):
        """Get range of key-value pairs

        Returns iterator of key-value pairs for keys in[low_key .. high_key]
        """
        # validate range
        if low_key is not None and high_key is not None and low_key >= high_key:
            return

        # find first leaf
        desc_to_node = self.provider.desc_to_node
        node = self.provider.root()
        if low_key is not None:
            for depth in range(self.provider.depth() - 1):
                node = desc_to_node(node.children[bisect(node.keys, low_key)])
            index = bisect_left(node.keys, low_key)
            if index >= len(node.keys):
                next = desc_to_node(node.next)
                if next is None:
                    return
                node, index = next, 0
        else:
            for depth in range(self.provider.depth() - 1):
                node = desc_to_node(node.children[0])
            index = 0

        # iterate over whole leafs
        while not high_key or node.keys[-1] < high_key:
            for index in range(index, len(node.keys)):
                yield node.keys[index], node.children[index]
            node = desc_to_node(node.next)
            if node is None:
                return
            index = 0

        # iterate over last leaf
        for index in range(index, len(node.keys)):
            key, value = node.keys[index], node.children[index]
            if key > high_key:
                return
            yield key, value

    def __set(self, key, value):
        """Associate key with value

        If key has already been, replace it with new value.
        """

        order = self.provider.order()
        dirty = self.provider.dirty
        desc_to_node = self.provider.desc_to_node
        node_to_desc = self.provider.node_to_desc

        # find path
        node, path = self.provider.root(), []
        for depth in range(self.provider.depth() - 1):
            index = bisect(node.keys, key)
            path.append((index, index + 1, node))
            node = desc_to_node(node.children[index])

        # check if value is updated
        index = bisect_left(node.keys, key)
        if index < len(node.keys) and key == node.keys[index]:
            node.children[index] = value
            dirty(node)
            return
        path.append((index, index, node))

        # size += 1
        self.provider.size(self.provider.size() + 1)

        # update tree
        sibling = None
        while path:
            key_index, child_index, node = path.pop()

            # add new key
            node.keys.insert(key_index, key)
            node.children.insert(child_index, value)
            dirty(node)

            if len(node.keys) < order:
                return

            # node is full so we need to split it
            center = len(node.children) >> 1
            keys, node.keys = node.keys[center:], node.keys[:center]
            children, node.children = node.children[center:], node.children[:center]

            if node.is_leaf:
                # create right sibling
                sibling = self.provider.create(keys, children, True)

                # keep leafs linked
                sibling_desc, node_next_desc = node_to_desc(sibling), node.next
                node.next, sibling.prev = sibling_desc, node_to_desc(node)
                node_next = desc_to_node(node_next_desc)
                if node_next:
                    node_next.prev, sibling.next = sibling_desc, node_next_desc
                    dirty(node_next)

                # update key
                key, value = sibling.keys[0], sibling_desc

            else:
                # create right sibling
                sibling = self.provider.create(keys, children, False)

                # update key
                key, value = node.keys.pop(), node_to_desc(sibling)

            dirty(sibling)

        # create new root
        self.provider.depth(self.provider.depth() + 1)  # depth += 1
        self.provider.root(self.provider.create(
                           [key],
                           [node_to_desc(self.provider.root()), node_to_desc(sibling)],
                           False))

    def __pop(self, key, value=nothing):
        """Pop value associated with key

        Remove value associated with key if any. Returns this value or value
        argument if it is set, else raises KeyValue exception.
        """

        half_order = self.provider.order() >> 1
        dirty = self.provider.dirty
        desc_to_node = self.provider.desc_to_node

        # find path
        node, path = self.provider.root(), []
        for depth in range(self.provider.depth() - 1):
            index = bisect(node.keys, key)
            parent, node = node, desc_to_node(node.children[index])
            path.append((node, index, parent))

        # check if key exists
        index = bisect_left(node.keys, key)
        if index >= len(node.keys) or key != node.keys[index]:
            if value is self.nothing:
                raise KeyError(key)
            return value
        value = node.children[index]
        key_index, child_index = index, index

        # size -= 1
        self.provider.size(self.provider.size() - 1)

        # update tree
        while path:
            node, node_index, parent = path.pop()

            # remove scheduled(key | child)
            del node.keys[key_index]
            del node.children[child_index]

            if len(node.keys) >= half_order:
                dirty(node)
                return value

            ## redistribute
            left, right = None, None
            if node_index > 0:
                # has left sibling
                left = desc_to_node(parent.children[node_index - 1])
                if len(left.keys) > half_order:  # borrow from left sibling
                    # copy correct key to node
                    node.keys.insert(0, left.keys[-1] if node.is_leaf
                                     else parent.keys[node_index - 1])
                    # move left key to parent
                    parent.keys[node_index - 1] = left.keys.pop()
                    # move left child to node
                    node.children.insert(0, left.children.pop())

                    dirty(node), dirty(left), dirty(parent)
                    return value

            if node_index < len(parent.keys):
                # has right sibling
                right = desc_to_node(parent.children[node_index + 1])
                if len(right.keys) > half_order:  # borrow from right sibling
                    if node.is_leaf:
                        # move right key to node
                        node.keys.append(right.keys.pop(0))
                        # copy next right key to parent
                        parent.keys[node_index] = right.keys[0]
                    else:
                        # copy correct key to node
                        node.keys.append(parent.keys[node_index])
                        # move right key to parent
                        parent.keys[node_index] = right.keys.pop(0)
                    # move right child to node
                    node.children.append(right.children.pop(0))

                    dirty(node), dirty(right), dirty(parent)
                    return value

            ## merge
            src, dst, child_index = ((node, left, node_index) if left
                                     else(right, node, node_index + 1))

            if node.is_leaf:
                # keep leafs linked
                dst.next = src.next
                src_next = desc_to_node(src.next)
                if src_next is not None:
                    src_next.prev = src.prev
                    dirty(src_next)
            else:
                # copy parents key
                dst.keys.append(parent.keys[child_index - 1])

            # copy node's(keys | children)
            dst.keys.extend(src.keys)
            dst.children.extend(src.children)

            # mark nodes
            self.provider.release(src)
            dirty(dst)

            # update key index
            key_index = child_index - 1

        ## update root
        root = self.provider.root()
        del root.keys[key_index]
        del root.children[child_index]

        if not root.keys:
            depth = self.provider.depth()
            if depth > 1:
                # root is not leaf because depth > 1
                self.provider.root(desc_to_node(*root.children))
                self.provider.release(root)
                self.provider.depth(depth - 1)  # depth -= 1
        else:
            dirty(root)

        return value

    def __len__(self):
        return self.provider.size()

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.__get_range(low_key=key.start, high_key=key.stop)
        return self.__get(key)

    def __setitem__(self, key, value):
        return self.__set(key, value)

    def __delitem__(self, key):
        return self.__pop(key)

    def __iter__(self):
        return map(itemgetter(0), self.__get_range())

    def __contains__(self, key):
        return self.__get(key, None) is not None

    def get(self, key, default=None):
        return self.__get(key, default)

    def pop(self, key, default=None):
        return self.__pop(key, default)

    def items(self):
        return self.__get_range()

    def values(self):
        return map(itemgetter(1), self.__get_range())


class BPTreeNode(object):
    """B+Tree Node
    """
    __slots__ = ('keys', 'children', 'is_leaf')

    def __init__(self, keys, children):
        self.keys = keys
        self.children = children
        self.is_leaf = False

    def __str__(self):
        return 'Node(keys:{}, children:{})'.format(self.keys, self.children)

    def __repr__(self):
        return str(self)


class BPTreeLeaf(BPTreeNode):
    """B+Tree Node
    """
    __slots__ = ('keys', 'children', 'prev', 'next', 'is_leaf')

    def __init__(self, keys, children):
        self.keys = keys
        self.children = children
        self.prev = None
        self.next = None
        self.is_leaf = True

    def __str__(self):
        return ('Leaf(prev:{}, next:{}, keys:{}, children:{})'
                .format(self.prev, self.next, self.keys, self.children))




__all__ = []


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import monad, async, do_green

    suite = TestSuite()
    for test in (monad, async, do_green):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite



import io
import random
import unittest
from ..map.bptree import BPTree
from ..map.provider.memory import MemoryBPTreeProvider
from ..map.provider.store import StoreBPTreeProvider
from ..store import StreamStore

__all__ = ('BPTreeTest',)


class BPTreeTest(unittest.TestCase):
    """B+Tree unit tests
    """
    def test_consist(self):
        """B+Tree consistency tests
        """
        provider = self.provider()
        tree, std = BPTree(provider), {}

        # compare tree and map mappings
        def validate(tree):
            self.assertEqual(len(tree), len(list(tree)))
            self.assertEqual(len(tree), len(std))
            for k, v in tree.items():
                self.assertEqual(std.get(k), v)
            for k, v in std.items():
                self.assertEqual(tree.get(k), v)

        # item count
        count = 1 << 10

        # reload
        provider = self.provider(provider)
        tree = BPTree(provider)
        validate(tree)

        # Insertion(10 .. 1024)
        for i in range(10, count):
            tree[i], std[i] = str(i), str(i)
        validate(tree)

        # reload
        provider = self.provider(provider)
        tree = BPTree(provider)
        validate(tree)

        # Insert(0 .. 10)
        for i in range(0, 10):
            tree[i], std[i] = str(i), str(i)
        validate(tree)

        # reload
        provider = self.provider(provider)
        tree = BPTree(provider)
        validate(tree)

        # Range
        self.assertEqual(list(tree[100:201]), [(key, str(key)) for key in range(100, 202)])
        self.assertEqual(list(tree[:100]), [(key, str(key)) for key in range(0, 101)])
        self.assertEqual(list(tree[101:102.5]), [(101, '101'), (102, '102')])
        self.assertEqual(list(tree[1022:]), [(1022, '1022'), (1023, '1023')])
        self.assertEqual(list(tree[100:10]), [])

        # Deletion
        keys = list(range(count))
        half = len(keys) >> 1
        random.shuffle(keys)
        for i in keys[:half]:
            self.assertEqual(tree.pop(i), std.pop(i))
        validate(tree)

        # reload
        provider = self.provider(provider)
        tree = BPTree(provider)
        validate(tree)

        for i in keys[half:]:
            self.assertEqual(tree.pop(i), std.pop(i))
        self.assertEqual(len(tree), 0)
        validate(tree)

        # reload
        provider = self.provider(provider)
        tree = BPTree(provider)
        validate(tree)

        return provider

    def provider(self, provider=None):
        """Memory B+Tree provider
        """
        if provider is None:
            return MemoryBPTreeProvider(order=7)
        return provider


class StoreBPTreeTest(BPTreeTest):
    """B+Tree with store provider unit tests
    """
    def provider(self, provider=None):
        """Store B+Tree provider
        """
        if provider is None:
            store = StreamStore(io.BytesIO())
        else:
            provider.flush()
            provider.store.flush()
            store = StreamStore(provider.store.stream)
        return StoreBPTreeProvider(store, store.create_cell('::test'), order=7)

    def test_consist(self):
        provider = BPTreeTest.test_consist(self)
        provider.drop()
        self.assertEqual(provider.store.size, 0)
